#!/bin/sh

if [ $# -lt 1 ]; then
    echo "error.. need args"
    exit 1
fi

function timens(){
    current=`date "+%Y%m%d%H%M%S%N"`
    echo $current
}

cpp=""
cmd=""
dir=""
cppname=""
outdir=""
schema=""

bindir=$(dirname ${BASH_SOURCE[0]})

for arg in "$@"
do
    if [ ${#arg} -gt 2 ]; then
        tmp=$(echo $arg | awk '/cpp$/||/cc$/||/cxx$/||/c\+\+$/{print $1}')
        if [ ${#tmp} -gt 2 ]; then
            cpp=$tmp
            dir=$(dirname $cpp)
            cppname=$(basename $cpp)
            curns=$(timens)
            outdir=$dir"/$curns/"
            cmd=$cmd" "$outdir$cppname
        else 
            cmd=$cmd" "$arg
        fi
    else
        cmd=$cmd" "$arg
    fi
done

if [ ${#cpp} -gt 2 ]; then
    schema=$(awk '/STORAGE_SCHEMA_BEGIN/,/STORAGE_SCHEMA_END/{print }' $cpp | awk 'NR>1{print p;p=$0}')
fi

if [ ${#schema} -gt 2 ];then
    mkdir -p $outdir
    outschema=$outdir"storage.fbs"
    echo "${schema}" > $outschema
    newcppfile=$outdir$cppname
    cp $cpp $newcppfile
    ${bindir}/myflatc.sh $outdir $outschema
    if [ $? -ne 0 ]; then
        rm -rf $outdir
        exit 1
    fi

    #Extract just the C++ code.
    sed -ne  "/STORAGE_SCHEMA_BEGIN/,/STORAGE_SCHEMA_END/!w $newcppfile.tmp" $newcppfile

    #Insert the header file. Keep the below format to maintain portable between Linux and MacOS.
    sed -i -e '/include/a\
               #include "storage_ant_generated.h"
              ' $newcppfile.tmp

    mv $newcppfile.tmp $newcppfile
    ${bindir}/my++ $cmd -I $outdir
    ret=$?
    if [ $ret -eq 0 ];then
        rm -rf $outdir
    fi
    exit $ret
else
    ${bindir}/my++ $@
    exit $?
fi

function rand(){  
    min=$1  
    max=$(($2-$min+1))  
    num=$(($RANDOM+1000000000))
    echo $(($num%$max+$min))  
}

