#ifndef _MATH_H
#define _MATH_H

#ifdef __cplusplus
extern "C" {
#endif

#include <features.h>

#define __NEED_float_t
#define __NEED_double_t
#include <bits/alltypes.h>

#if 100*__GNUC__+__GNUC_MINOR__ >= 303
#define NAN       __builtin_nanf("")
#define INFINITY  __builtin_inff()
#else
#define NAN       (0.0f/0.0f)
#define INFINITY  1e5000f
#endif

#define HUGE_VALF INFINITY
#define HUGE_VAL  ((double)INFINITY)
#define HUGE_VALL ((long_double)INFINITY)

#define MATH_ERRNO  1
#define MATH_ERREXCEPT 2
#define math_errhandling 2

#define FP_ILOGBNAN (-1-(int)(((unsigned)-1)>>1))
#define FP_ILOGB0 FP_ILOGBNAN

#define FP_NAN       0
#define FP_INFINITE  1
#define FP_ZERO      2
#define FP_SUBNORMAL 3
#define FP_NORMAL    4

int __fpclassify(double);
int __fpclassifyf(float);
int __fpclassifyl(long_double);

static __inline unsigned __FLOAT_BITS(float __f)
{
	union {float __f; unsigned __i;} __u;
	__u.__f = __f;
	return __u.__i;
}
static __inline unsigned long long __DOUBLE_BITS(double __f)
{
	union {double __f; unsigned long long __i;} __u;
	__u.__f = __f;
	return __u.__i;
}

#define fpclassify(x) ( \
	sizeof(x) == sizeof(float) ? __fpclassifyf(x) : \
	sizeof(x) == sizeof(double) ? __fpclassify(x) : \
	__fpclassifyl(x) )

#define isinf(x) ( \
	sizeof(x) == sizeof(float) ? (__FLOAT_BITS(x) & 0x7fffffff) == 0x7f800000 : \
	sizeof(x) == sizeof(double) ? (__DOUBLE_BITS(x) & -1ULL>>1) == 0x7ffULL<<52 : \
	__fpclassifyl(x) == FP_INFINITE)

#define isnan(x) ( \
	sizeof(x) == sizeof(float) ? (__FLOAT_BITS(x) & 0x7fffffff) > 0x7f800000 : \
	sizeof(x) == sizeof(double) ? (__DOUBLE_BITS(x) & -1ULL>>1) > 0x7ffULL<<52 : \
	__fpclassifyl(x) == FP_NAN)

#define isnormal(x) ( \
	sizeof(x) == sizeof(float) ? ((__FLOAT_BITS(x)+0x00800000) & 0x7fffffff) >= 0x01000000 : \
	sizeof(x) == sizeof(double) ? ((__DOUBLE_BITS(x)+(1ULL<<52)) & -1ULL>>1) >= 1ULL<<53 : \
	__fpclassifyl(x) == FP_NORMAL)

#define isfinite(x) ( \
	sizeof(x) == sizeof(float) ? (__FLOAT_BITS(x) & 0x7fffffff) < 0x7f800000 : \
	sizeof(x) == sizeof(double) ? (__DOUBLE_BITS(x) & -1ULL>>1) < 0x7ffULL<<52 : \
	__fpclassifyl(x) > FP_INFINITE)

int __signbit(double);
int __signbitf(float);
int __signbitl(long_double);

#define signbit(x) ( \
	sizeof(x) == sizeof(float) ? (int)(__FLOAT_BITS(x)>>31) : \
	sizeof(x) == sizeof(double) ? (int)(__DOUBLE_BITS(x)>>63) : \
	__signbitl(x) )

#define isunordered(x,y) (isnan((x)) ? ((void)(y),1) : isnan((y)))

#define __ISREL_DEF(rel, op, type) \
static __inline int __is##rel(type __x, type __y) \
{ return !isunordered(__x,__y) && __x op __y; }

__ISREL_DEF(lessf, <, float_t)
__ISREL_DEF(less, <, double_t)
__ISREL_DEF(lessl, <, long_double)
__ISREL_DEF(lessequalf, <=, float_t)
__ISREL_DEF(lessequal, <=, double_t)
__ISREL_DEF(lessequall, <=, long_double)
__ISREL_DEF(lessgreaterf, !=, float_t)
__ISREL_DEF(lessgreater, !=, double_t)
__ISREL_DEF(lessgreaterl, !=, long_double)
__ISREL_DEF(greaterf, >, float_t)
__ISREL_DEF(greater, >, double_t)
__ISREL_DEF(greaterl, >, long_double)
__ISREL_DEF(greaterequalf, >=, float_t)
__ISREL_DEF(greaterequal, >=, double_t)
__ISREL_DEF(greaterequall, >=, long_double)

#define __tg_pred_2(x, y, p) ( \
	sizeof((x)+(y)) == sizeof(float) ? p##f(x, y) : \
	sizeof((x)+(y)) == sizeof(double) ? p(x, y) : \
	p##l(x, y) )

#define isless(x, y)            __tg_pred_2(x, y, __isless)
#define islessequal(x, y)       __tg_pred_2(x, y, __islessequal)
#define islessgreater(x, y)     __tg_pred_2(x, y, __islessgreater)
#define isgreater(x, y)         __tg_pred_2(x, y, __isgreater)
#define isgreaterequal(x, y)    __tg_pred_2(x, y, __isgreaterequal)

double      acos(double);
float       acosf(float);
long_double acosl(long_double);

double      acosh(double);
float       acoshf(float);
long_double acoshl(long_double);

double      asin(double);
float       asinf(float);
long_double asinl(long_double);

double      asinh(double);
float       asinhf(float);
long_double asinhl(long_double);

double      atan(double);
float       atanf(float);
long_double atanl(long_double);

double      atan2(double, double);
float       atan2f(float, float);
long_double atan2l(long_double, long_double);

double      atanh(double);
float       atanhf(float);
long_double atanhl(long_double);

double      cbrt(double);
float       cbrtf(float);
long_double cbrtl(long_double);

double      ceil(double);
float       ceilf(float);
long_double ceill(long_double);

double      copysign(double, double);
float       copysignf(float, float);
long_double copysignl(long_double, long_double);

double      cos(double);
float       cosf(float);
long_double cosl(long_double);

double      cosh(double);
float       coshf(float);
long_double coshl(long_double);

double      erf(double);
float       erff(float);
long_double erfl(long_double);

double      erfc(double);
float       erfcf(float);
long_double erfcl(long_double);

double      exp(double);
float       expf(float);
long_double expl(long_double);

double      exp2(double);
float       exp2f(float);
long_double exp2l(long_double);

double      expm1(double);
float       expm1f(float);
long_double expm1l(long_double);

double      fabs(double);
float       fabsf(float);
long_double fabsl(long_double);

double      fdim(double, double);
float       fdimf(float, float);
long_double fdiml(long_double, long_double);

double      floor(double);
float       floorf(float);
long_double floorl(long_double);

double      fma(double, double, double);
   //float       fmaf(float, float, float);
   //long_double fmal(long_double, long_double, long_double);

double      fmax(double, double);
float       fmaxf(float, float);
long_double fmaxl(long_double, long_double);

double      fmin(double, double);
float       fminf(float, float);
long_double fminl(long_double, long_double);

double      fmod(double, double);
float       fmodf(float, float);
long_double fmodl(long_double, long_double);

double      frexp(double, int *);
float       frexpf(float, int *);
long_double frexpl(long_double, int *);

double      hypot(double, double);
float       hypotf(float, float);
long_double hypotl(long_double, long_double);

int         ilogb(double);
int         ilogbf(float);
int         ilogbl(long_double);

double      ldexp(double, int);
float       ldexpf(float, int);
long_double ldexpl(long_double, int);

double      lgamma(double);
float       lgammaf(float);
long_double lgammal(long_double);

long long   llrint(double);
long long   llrintf(float);
long long   llrintl(long_double);

long long   llround(double);
long long   llroundf(float);
long long   llroundl(long_double);

double      log(double);
float       logf(float);
long_double logl(long_double);

double      log10(double);
float       log10f(float);
long_double log10l(long_double);

double      log1p(double);
float       log1pf(float);
long_double log1pl(long_double);

double      log2(double);
float       log2f(float);
long_double log2l(long_double);

double      logb(double);
float       logbf(float);
long_double logbl(long_double);

long        lrint(double);
long        lrintf(float);
long        lrintl(long_double);

long        lround(double);
long        lroundf(float);
long        lroundl(long_double);

double      modf(double, double *);
float       modff(float, float *);
long_double modfl(long_double, long_double *);

double      nan(const char *);
float       nanf(const char *);
long_double nanl(const char *);

double      nearbyint(double);
float       nearbyintf(float);
long_double nearbyintl(long_double);

double      nextafter(double, double);
float       nextafterf(float, float);
long_double nextafterl(long_double, long_double);

double      nexttoward(double, long_double);
float       nexttowardf(float, long_double);
long_double nexttowardl(long_double, long_double);

double      pow(double, double);
float       powf(float, float);
long_double powl(long_double, long_double);

double      remainder(double, double);
float       remainderf(float, float);
long_double remainderl(long_double, long_double);

double      remquo(double, double, int *);
float       remquof(float, float, int *);
long_double remquol(long_double, long_double, int *);

double      rint(double);
float       rintf(float);
long_double rintl(long_double);

double      round(double);
float       roundf(float);
long_double roundl(long_double);

double      scalbln(double, long);
float       scalblnf(float, long);
long_double scalblnl(long_double, long);

double      scalbn(double, int);
float       scalbnf(float, int);
long_double scalbnl(long_double, int);

double      sin(double);
float       sinf(float);
long_double sinl(long_double);

double      sinh(double);
float       sinhf(float);
long_double sinhl(long_double);

double      sqrt(double);
float       sqrtf(float);
long_double sqrtl(long_double);

double      tan(double);
float       tanf(float);
long_double tanl(long_double);

double      tanh(double);
float       tanhf(float);
long_double tanhl(long_double);

double      tgamma(double);
float       tgammaf(float);
long_double tgammal(long_double);

double      trunc(double);
float       truncf(float);
long_double truncl(long_double);


#if defined(_XOPEN_SOURCE) || defined(_BSD_SOURCE)
#undef  MAXFLOAT
#define MAXFLOAT        3.40282346638528859812e+38F
#endif

#if defined(_XOPEN_SOURCE) || defined(_GNU_SOURCE) || defined(_BSD_SOURCE)
#define M_E             2.7182818284590452354   /* e */
#define M_LOG2E         1.4426950408889634074   /* log_2 e */
#define M_LOG10E        0.43429448190325182765  /* log_10 e */
#define M_LN2           0.69314718055994530942  /* log_e 2 */
#define M_LN10          2.30258509299404568402  /* log_e 10 */
#define M_PI            3.14159265358979323846  /* pi */
#define M_PI_2          1.57079632679489661923  /* pi/2 */
#define M_PI_4          0.78539816339744830962  /* pi/4 */
#define M_1_PI          0.31830988618379067154  /* 1/pi */
#define M_2_PI          0.63661977236758134308  /* 2/pi */
#define M_2_SQRTPI      1.12837916709551257390  /* 2/sqrt(pi) */
#define M_SQRT2         1.41421356237309504880  /* sqrt(2) */
#define M_SQRT1_2       0.70710678118654752440  /* 1/sqrt(2) */

extern int signgam;

double      j0(double);
double      j1(double);
double      jn(int, double);

double      y0(double);
double      y1(double);
double      yn(int, double);
#endif

#if defined(_GNU_SOURCE) || defined(_BSD_SOURCE)
#define HUGE            3.40282346638528859812e+38F

double      drem(double, double);
float       dremf(float, float);

int         finite(double);
int         finitef(float);

double      scalb(double, double);
float       scalbf(float, float);

double      significand(double);
float       significandf(float);

double      lgamma_r(double, int*);
float       lgammaf_r(float, int*);

float       j0f(float);
float       j1f(float);
float       jnf(int, float);

float       y0f(float);
float       y1f(float);
float       ynf(int, float);
#endif

#ifdef _GNU_SOURCE
long_double lgammal_r(long_double, int*);

void        sincos(double, double*, double*);
void        sincosf(float, float*, float*);
void        sincosl(long_double, long_double*, long_double*);

double      exp10(double);
float       exp10f(float);
long_double exp10l(long_double);

double      pow10(double);
float       pow10f(float);
long_double pow10l(long_double);
#endif

#ifdef __cplusplus
}
#endif

#endif
