
#ifndef MYBUFFER_ALDABALIB_H
#define MYBUFFER_ALDABALIB_H

#include <cstdarg>
#include <string>

#include "aldaba/aldaba.h"  // Provided by aldaba cdt
#include "mybuffer/mychainlib.h"

class AldabaLib : public MychainLib {
public:
    virtual ~AldabaLib() = default;

    int MyPrint(const char *format, ...) override {
        char buf[LOGGER_BUFFER_SIZE];
        va_list args;
        va_start(args, format);
        vsnprintf(buf, LOGGER_BUFFER_SIZE, format, args);
        va_end(args);
        return aldaba::println("%s", buf);
    }

    int MyPrintWithLevel(LogLevel level, const char* format, ...) override {
        if (level >= MYBUFFER_LOG_LEVEL) {
            char buf[LOGGER_BUFFER_SIZE];
            va_list args;
            va_start(args, format);
            vsnprintf(buf, LOGGER_BUFFER_SIZE, (kLevelPrefixMap[level] + format).c_str(), args);
            va_end(args);
            return aldaba::println("%s", buf);
        }

        return 0;
    }

    void MyRevert(const std::string &msg) override { aldaba::abort(msg); }

    int MyGetStorage(const std::string &key, std::string &val) override {
        int rv = MyGetStorageInternal(key, val);
        if (rv != AldabaApiErrors::ALDABA_ERROR_NONE) {
            if (rv != AldabaApiErrors::ALDABA_ERROR_KEY_NOT_EXISTS) {
                std::string msg = "Internal get-kv error: " + std::to_string(rv);
                MYLOG(LogLevel::kError, msg);
                MyRevert(msg);
            }
            else {
                MYLOG(LogLevel::kWarning, "Failed to get storage for key %s error %d", key.c_str(), rv);
            }
        }
        return rv;
    }

    int MySetStorage(const std::string &key, const std::string &val) override {
        int rv = MySetStorageInternal(key, val);
        if (rv != AldabaApiErrors::ALDABA_ERROR_NONE) {
            std::string msg = "Internal set-kv error: " + std::to_string(rv);
            MYLOG(LogLevel::kError, msg);
            MyRevert(msg);
        }
        return rv;
    }

    int MyDelStorage(const std::string &key) override {
        int rv = MyDelStorageInternal(key);
        if (rv != AldabaApiErrors::ALDABA_ERROR_NONE) {
            std::string msg = "Internal del-kv error: " + std::to_string(rv);
            MYLOG(LogLevel::kError, msg);
            MyRevert(msg);
        }
        return rv;
    }
};

#endif // MYBUFFER_ALDABALIB_H
