#ifndef MYBUFFER_ALDABA_MAPITERABLE_H
#define MYBUFFER_ALDABA_MAPITERABLE_H

#include <third_party/cityhash/city.h>

#include <deque>
#include <map>
#include <memory>

#include "flatbuffers/flatbuffers.h"
#include "mybuffer/key.h"
#include "mybuffer/mybuffer_aldaba_metadata_generated.h"
#include "mybuffer/mybuffer_const_define.h"
#include "mybuffer/mybuffer_mybase.h"
#include "mybuffer/mybuffer_myfactory.h"
#include "mybuffer/mybuffer_mymap.h"
#include "mybuffer/mybuffer_mytable.h"
#include "mybuffer/mybuffer_myvector.h"
#include "mybuffer/mychainlib.h"

// M-level code of mybuffer_aldaba_metadata.fbs begins /////////////////////////

struct AlStringM : public MyTable<AlStringT, AlString> {
 private:
  using MyTable<AlStringT, AlString>::MyTable;
  friend class MyFactory<AlStringM, AlStringT>;
  friend class MyFactory<AlStringM>;

 public:
  ~AlStringM() {}
  std::string get_val() { return native_table_ptr_->val; }
  bool set_val(std::string _val) {
    uint32_t size_limit = 0 ? 0 : kDefaultStringSizeLimit;
    if (_val.length() > size_limit) {
      std::string msg = "The string size is over the limit" + std::to_string(size_limit);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    set_dirty(true);
    native_table_ptr_->val = std::move(_val);
    return true;
  }
};

struct AlUint32M : public MyTable<AlUint32T, AlUint32> {
 private:
  using MyTable<AlUint32T, AlUint32>::MyTable;
  friend class MyFactory<AlUint32M, AlUint32T>;
  friend class MyFactory<AlUint32M>;

 public:
  ~AlUint32M() {}
  uint32_t get_val() { return native_table_ptr_->val; }
  bool set_val(uint32_t _val) {
    set_dirty(true);
    native_table_ptr_->val = _val;
    return true;
  }
};

struct AggregatedValueM : public MyTable<AggregatedValueT, AggregatedValue> {
 private:
  using MyTable<AggregatedValueT, AggregatedValue>::MyTable;
  friend class MyFactory<AggregatedValueM, AggregatedValueT>;
  friend class MyFactory<AggregatedValueM>;

 public:
  ~AggregatedValueM() {}
  std::string get_buffer() { return native_table_ptr_->buffer; }
  bool set_buffer(std::string _buffer) {
    uint32_t size_limit = 0 ? 0 : kDefaultStringSizeLimit;
    if (_buffer.length() > size_limit) {
      std::string msg = "The string size is over the limit" + std::to_string(size_limit);
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }
    set_dirty(true);
    native_table_ptr_->buffer = std::move(_buffer);
    return true;
  }
};

struct SpreadedValueM : public MyTable<SpreadedValueT, SpreadedValue,
                                       true /*as_pass_through_buffer*/> {
 private:
  using MyTable<SpreadedValueT, SpreadedValue,
                true /*as_pass_through_buffer*/>::MyTable;
  friend class MyFactory<SpreadedValueM, SpreadedValueT>;
  friend class MyFactory<SpreadedValueM>;

 public:
  ~SpreadedValueM() {}
  std::string get_buffer() { return pass_through_buffer_; }
  bool set_buffer(std::string buffer) {
    set_dirty(true);
    pass_through_buffer_ = buffer;
    return true;
  }
};

struct AlIndexNodeM : public MyTable<AlIndexNodeT, AlIndexNode> {
 private:
  typedef MyVector<uint32_t, uint32_t, true> VectorOfUint32;
  std::weak_ptr<MyVector<AlStringM, AlStringT>> keys_;
  std::weak_ptr<MyVector<AlUint32M, AlUint32T>> bitmap_;
  std::weak_ptr<MyVector<AggregatedValueM, AggregatedValueT>> values_;
  std::weak_ptr<VectorOfUint32> bitmap_new_;
  using MyTable<AlIndexNodeT, AlIndexNode>::MyTable;
  friend class MyFactory<AlIndexNodeM, AlIndexNodeT>;
  friend class MyFactory<AlIndexNodeM>;

 public:
  ~AlIndexNodeM() {}
  uint16_t get_lcount() { return native_table_ptr_->lcount; }
  bool set_lcount(uint16_t _lcount) {
    set_dirty(true);
    native_table_ptr_->lcount = _lcount;
    return true;
  }
  uint64_t get_tcount() { return native_table_ptr_->tcount; }
  bool set_tcount(uint64_t _tcount) {
    set_dirty(true);
    native_table_ptr_->tcount = _tcount;
    return true;
  }
  std::shared_ptr<MyVector<AlStringM, AlStringT>> get_keys() {
    std::shared_ptr<MyVector<AlStringM, AlStringT>> rv;
    rv = keys_.lock();
    if (!rv) {
      rv = MyFactory<MyVector<AlStringM, AlStringT>, AlStringT>::make_instance(
          mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::VECTOR, 2 /*field_id*/),
          native_table_ptr_->keys, get_self_ptr(), 0);
      keys_ = rv;
    }
    return rv;
  }
  std::shared_ptr<MyVector<AlUint32M, AlUint32T>> get_bitmap() {
    std::shared_ptr<MyVector<AlUint32M, AlUint32T>> rv;
    rv = bitmap_.lock();
    if (!rv) {
      rv = MyFactory<MyVector<AlUint32M, AlUint32T>, AlUint32T>::make_instance(
          mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::VECTOR, 3 /*field_id*/),
          native_table_ptr_->bitmap, get_self_ptr(), 0);
      bitmap_ = rv;
    }
    return rv;
  }
  std::shared_ptr<MyVector<AggregatedValueM, AggregatedValueT>> get_values() {
    std::shared_ptr<MyVector<AggregatedValueM, AggregatedValueT>> rv;
    rv = values_.lock();
    if (!rv) {
      rv = MyFactory<
          MyVector<AggregatedValueM, AggregatedValueT>,
          AggregatedValueT>::make_instance(mychain_lib_ptr_,
                                           base_key_->get_table_field_key(
                                               FieldNodeType::VECTOR,
                                               4 /*field_id*/),
                                           native_table_ptr_->values,
                                           get_self_ptr(), 0);
      values_ = rv;
    }
    return rv;
  }
  std::shared_ptr<VectorOfUint32> get_bitmap_new() {
    std::shared_ptr<VectorOfUint32> rv;
    rv = bitmap_new_.lock();
    if (!rv) {
      rv = MyFactory<VectorOfUint32, uint32_t>::make_instance(
          mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::VECTOR, 5 /*field_id*/),
          native_table_ptr_->bitmap_new, get_self_ptr(), 0);
      bitmap_new_ = rv;
    }
    return rv;
  }
};

struct AlMetaM : public MyTable<AlMetaT, AlMeta> {
 private:
  typedef MyVector<uint32_t, uint32_t, true> VectorOfUint32;
  std::weak_ptr<MyVector<AlUint32M, AlUint32T>> bitmap_;
  std::weak_ptr<VectorOfUint32> bitmap_new_;
  using MyTable<AlMetaT, AlMeta>::MyTable;
  friend class MyFactory<AlMetaM, AlMetaT>;
  friend class MyFactory<AlMetaM>;

 public:
  ~AlMetaM() {}
  HashFunc get_hash_function() { return native_table_ptr_->hash_function; }
  bool set_hash_function(HashFunc _hash_function) {
    set_dirty(true);
    native_table_ptr_->hash_function = _hash_function;
    return true;
  }
  uint16_t get_index_width() { return native_table_ptr_->index_width; }
  bool set_index_width(uint16_t _index_width) {
    set_dirty(true);
    native_table_ptr_->index_width = _index_width;
    return true;
  }
  uint16_t get_max_lcount() { return native_table_ptr_->max_lcount; }
  bool set_max_lcount(uint16_t _max_lcount) {
    set_dirty(true);
    native_table_ptr_->max_lcount = _max_lcount;
    return true;
  }
  AlStorageLayout get_storage_layout() {
    return native_table_ptr_->storage_layout;
  }
  bool set_storage_layout(AlStorageLayout _storage_layout) {
    set_dirty(true);
    native_table_ptr_->storage_layout = _storage_layout;
    return true;
  }
  std::shared_ptr<MyVector<AlUint32M, AlUint32T>> get_bitmap() {
    std::shared_ptr<MyVector<AlUint32M, AlUint32T>> rv;
    rv = bitmap_.lock();
    if (!rv) {
      rv = MyFactory<MyVector<AlUint32M, AlUint32T>, AlUint32T>::make_instance(
          mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::VECTOR, 4 /*field_id*/),
          native_table_ptr_->bitmap, get_self_ptr(), 0);
      bitmap_ = rv;
    }
    return rv;
  }
  std::shared_ptr<VectorOfUint32> get_bitmap_new() {
    std::shared_ptr<VectorOfUint32> rv;
    rv = bitmap_new_.lock();
    if (!rv) {
      rv = MyFactory<VectorOfUint32, uint32_t>::make_instance(
          mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::VECTOR, 5 /*field_id*/),
          native_table_ptr_->bitmap_new, get_self_ptr(), 0);
      bitmap_new_ = rv;
    }
    return rv;
  }
};

// M-level code of mybuffer_aldaba_metadata.fbs ends ///////////////////////////

static inline bool IsAggregatedKVLayout(AlStorageLayout sl) {
  return sl == AlStorageLayout::AlStorageLayout_AggregatedKVLayout;
}

static inline bool IsAggregatedKeyLayout(AlStorageLayout sl) {
  return sl == AlStorageLayout::AlStorageLayout_AggregatedKeyLayout;
}

template <typename BitmapOwnerType = AlIndexNodeM>
class BitmapAccessor {
  typedef std::shared_ptr<MyVector<AlUint32M, AlUint32T>> OldBitmap;
  typedef std::shared_ptr<MyVector<uint32_t, uint32_t, true>> NewBitmap;
  typedef std::shared_ptr<BitmapOwnerType> BitmapOwner;

 public:
  static void print_bitmap(const BitmapOwner& bitmap_owner,
                           std::shared_ptr<MychainLib> mychain_lib_ptr) {
    std::string name = bitmap_owner->get_base_key();

    mychain_lib_ptr->MyPrint("Print new bitmap of %s: \n", name.data());
    auto bitmap_new = bitmap_owner->get_bitmap_new();
    for (uint32_t i = 0; i < bitmap_new->size(); i++) {
      mychain_lib_ptr->MyPrint("bitmap_new: %d\n", get_ith_val(i, bitmap_new));
    }

    mychain_lib_ptr->MyPrint("\tPrint old bitmap of %s: \n", name.data());
    auto bitmap = bitmap_owner->get_bitmap();
    for (uint32_t i = 0; i < bitmap->size(); i++) {
      mychain_lib_ptr->MyPrint("bitmap: %d\n", get_ith_val(i, bitmap));
    }

    return;
  }

  static bool read_bitmap(uint16_t index_id, const BitmapOwner& bitmap_owner,
                          uint16_t expected_size, const std::string& situation,
                          std::shared_ptr<MychainLib> mychain_lib_ptr) {
    revert_if_invalid_bitmap_size(bitmap_owner, expected_size, situation,
                                  mychain_lib_ptr);

    auto bitmap = bitmap_owner->get_bitmap();
    auto bitmap_new = bitmap_owner->get_bitmap_new();
    bool result = false;

    if (bitmap->size() != 0) {
      result = read_bitmap_<OldBitmap>(index_id, bitmap);
    } else if (bitmap_new->size() != 0) {
      result = read_bitmap_<NewBitmap>(index_id, bitmap_new);
    } else {
      // Never reach here
      std::string msg = "Invalid bitmap size when " + situation;
      mychain_lib_ptr->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr->MyRevert(msg);
    }
    return result;
  }

  static void write_bitmap(uint16_t index_id, bool bit,
                           const BitmapOwner& bitmap_owner,
                           uint16_t expected_size, const std::string& situation,
                           std::shared_ptr<MychainLib> mychain_lib_ptr) {
    revert_if_invalid_bitmap_size(bitmap_owner, expected_size, situation,
                                  mychain_lib_ptr);

    auto bitmap = bitmap_owner->get_bitmap();
    auto bitmap_new = bitmap_owner->get_bitmap_new();

    if (use_new_bitmap(mychain_lib_ptr)) {
      copy_and_clear_if_non_empty<OldBitmap, NewBitmap>(bitmap, bitmap_new);
      write_bitmap_<NewBitmap>(index_id, bit, bitmap_new);
    } else {
      copy_and_clear_if_non_empty<NewBitmap, OldBitmap>(bitmap_new, bitmap);
      write_bitmap_<OldBitmap>(index_id, bit, bitmap);
    }
    return;
  }

  static void fill_bitmap_with_zeros(
      uint16_t bitmap_size, const BitmapOwner& bitmap_owner,
      const std::shared_ptr<MychainLib>& mychain_lib_ptr) {
    if (use_new_bitmap(mychain_lib_ptr)) {
      auto bitmap_new = bitmap_owner->get_bitmap_new();
      if (bitmap_new->size() == 0) {
        for (uint16_t i = 0; i < bitmap_size; i++) {
          bitmap_new->append_element(0);
        }
      }
    } else {
      auto bitmap = bitmap_owner->get_bitmap();
      if (bitmap->size() == 0) {
        for (uint16_t i = 0; i < bitmap_size; i++) {
          auto pbits = bitmap->append_element();
          pbits->set_val(0);
        }
      }
    }
    return;
  }

  /**
   * @brief revert_if_invalid_bitmap_size checks the validity of size of bitmaps
   * The invariant here is that:
   * 1. there should be one and only one non-empty bitmap, with expected size;
   * 2. the other one should be empty;
   */
  static void revert_if_invalid_bitmap_size(
      const BitmapOwner& bitmap_owner, uint16_t expected_size,
      const std::string& situation,
      const std::shared_ptr<MychainLib>& mychain_lib_ptr) {
    auto bitmap = bitmap_owner->get_bitmap();
    auto bitmap_new = bitmap_owner->get_bitmap_new();

    if (!((bitmap->size() == expected_size && bitmap_new->size() == 0) ||
          (bitmap->size() == 0 && bitmap_new->size() == expected_size))) {
      std::string msg = "Invalid bitmap size when " + situation +
                        " got bitmap size: " + std::to_string(bitmap->size()) +
                        " got new bitmap size: " + std::to_string(bitmap_new->size()) +
                        " want bitmap size: " + std::to_string(expected_size);
      mychain_lib_ptr->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr->MyRevert(msg);
    }
    return;
  }

 private:
  static bool use_new_bitmap(
      const std::shared_ptr<MychainLib>& mychain_lib_ptr) {
    return (mychain_lib_ptr->GetCurrentBlockVersion() >=
            mychain_lib_ptr->GetBlockVersionOfReleaseTag("1.5.3"));
  }

  template <typename S, typename D>
  static void copy_and_clear_if_non_empty(const S& src, const D& dst) {
    if (src->size() != 0) {
      for (uint32_t i = 0; i < src->size(); i++) {
        append_val(get_ith_val(i, src), dst);
      }
      src->clear();
    }
    return;
  }

  static uint32_t get_ith_val(uint32_t idx, const OldBitmap& bitmap) {
    return bitmap->get_element(idx)->get_val();
  }
  static uint32_t get_ith_val(uint32_t idx, const NewBitmap& bitmap) {
    return bitmap->get_element(idx);
  }
  static void set_ith_val(uint32_t idx, uint32_t val, const OldBitmap& bitmap) {
    bitmap->get_element(idx)->set_val(val);
    return;
  }
  static void set_ith_val(uint32_t idx, uint32_t val, const NewBitmap& bitmap) {
    bitmap->set_element(idx, val);
    return;
  }
  static void append_val(uint32_t val, const OldBitmap& bitmap) {
    bitmap->append_element()->set_val(val);
    return;
  }
  static void append_val(uint32_t val, const NewBitmap& bitmap) {
    bitmap->append_element(val);
    return;
  }

  template <typename BitmapType>
  static bool read_bitmap_(uint16_t index_id, const BitmapType& bitmap) {
    uint16_t move = (31 - index_id % 32);
    uint16_t pos = index_id / 32;
    uint32_t bitmask = 1 << move;
    return (bool)((get_ith_val(pos, bitmap) & bitmask) >> (move));
  }

  template <typename BitmapType>
  static void write_bitmap_(uint16_t index_id, bool bit,
                            const BitmapType& bitmap) {
    uint16_t move = (31 - index_id % 32);
    uint16_t pos = index_id / 32;
    uint32_t bitmask = bit ? (1 << move) : (~(1 << move));
    uint32_t val = get_ith_val(pos, bitmap);
    val = bit ? (val | bitmask) : (val & bitmask);
    set_ith_val(pos, val, bitmap);
    return;
  }
};

template <class TMap, class M, AlStorageLayout storage_layout>
class IterItem;

template <class TMap, class M, AlStorageLayout storage_layout>
class AlMapKeyIterator {
 public:
  AlMapKeyIterator(const std::shared_ptr<TMap>& _map_ptr,
                   const std::shared_ptr<MychainLib>& _mychain_lib_ptr)
      : map_ptr_(_map_ptr), mychain_lib_ptr_(_mychain_lib_ptr) {
    Initialize();
  }

  AlMapKeyIterator(const AlMapKeyIterator& m) {
    map_ptr_ = m.map_ptr_;
    mychain_lib_ptr_ = m.mychain_lib_ptr_;
    Initialize();
  }

  ~AlMapKeyIterator() {}

  bool valid() {
    std::shared_ptr<AlIndexNodeM> pindex;
    return isValid_ && pnode_ && (pindex = pnode_->index_node_ptr_) &&
           (offset_ < pindex->get_lcount());
  }

  IterItem<TMap, M, storage_layout> operator*() {
    if (!valid()) {
      std::string msg = "The iterator is invalid.";
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
    }

    std::shared_ptr<AlIndexNodeM> pindex = pnode_->index_node_ptr_;
    std::shared_ptr<AlStringM> pkey = pindex->get_keys()->get_element(offset_);
    std::string key = pkey->get_val();
    std::shared_ptr<MyBase> pvalue;

    if (IsAggregatedKVLayout(storage_layout)) {
      pvalue = pindex->get_values()->get_element(offset_);
      return IterItem<TMap, M, storage_layout>(
          key, std::static_pointer_cast<M>(pvalue));
    } else if (IsAggregatedKeyLayout(storage_layout)) {
      return IterItem<TMap, M, storage_layout>(key, map_ptr_);
    } else {
      std::string msg = "The storage_layout is invalid.";
      mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      mychain_lib_ptr_->MyRevert(msg);
      // Never reach here
      return IterItem<TMap, M, storage_layout>(key, map_ptr_);
    }
  }

  AlMapKeyIterator<TMap, M, storage_layout>& operator++() {
    std::shared_ptr<AlIndexNodeM> pindex;
    isValid_ = false;
    if (pnode_ && (pindex = pnode_->index_node_ptr_) &&
        (offset_ < pindex->get_lcount() - 1)) {
      offset_++;
      isValid_ = true;
    } else {
      while (!nodes_.empty()) {
        pnode_ = std::move(nodes_.front());
        nodes_.pop_front();

        // Enqueue children tree nodes
        if ((pindex = pnode_->index_node_ptr_)) {
          for (uint16_t i = 0; i < map_ptr_->get_index_width(); i++) {
            if (BitmapAccessor<AlIndexNodeM>::read_bitmap(
                    i, pindex, map_ptr_->get_bitmap_size(), "iterator++",
                    mychain_lib_ptr_)) {
              std::string index_key = pindex->get_base_key().substr(
                  map_ptr_->get_base_key().size() + 1);
              std::shared_ptr<AlIndexNodeM> pidx =
                  map_ptr_->get_index_node_by_index_key(
                      Key::ConcatWithSlash(index_key, std::to_string(i)));
              if (pidx) {
                nodes_.push_back(std::unique_ptr<Node>(new Node(pidx)));
              }
            }
          }
        }

        if ((pindex = pnode_->index_node_ptr_) && pindex->get_lcount() > 0) {
          offset_ = 0;
          isValid_ = true;
          break;
        }
      }
    }
    return *this;
  }

 private:
  void Initialize() {
    if (initialized_) return;

    std::shared_ptr<AlMetaM> pmeta =
        map_ptr_->get_meta_data(false /*revert_on_non_existence*/);
    // If there is no meta, valid() will return false.
    if (!pmeta) return;

    for (uint16_t i = 0; i < map_ptr_->get_index_width(); i++) {
      if (BitmapAccessor<AlMetaM>::read_bitmap(
              i, pmeta, map_ptr_->get_bitmap_size(), "iterator initialization",
              mychain_lib_ptr_)) {
        std::shared_ptr<AlIndexNodeM> pindex =
            map_ptr_->get_index_node_by_index_key(std::to_string(i));
        if (pindex) {
          nodes_.push_back(std::unique_ptr<Node>(new Node(pindex)));
        }
      }
    }
    ++(*this);

    initialized_ = true;
  }

  struct Node {
    std::shared_ptr<AlIndexNodeM> index_node_ptr_;

    explicit Node(std::shared_ptr<AlIndexNodeM> _ptr) {
      index_node_ptr_ = _ptr;
    }

    ~Node() {
      if (index_node_ptr_) index_node_ptr_.reset();
    }
  };

  std::shared_ptr<TMap> map_ptr_;
  std::deque<std::unique_ptr<Node>> nodes_;
  std::unique_ptr<Node> pnode_;
  int offset_ = -1;
  bool isValid_ = false;
  bool initialized_ = false;
  std::shared_ptr<MychainLib> mychain_lib_ptr_;
};

template <class TMap, class M, AlStorageLayout storage_layout>
class IterItem {
 public:
  ~IterItem() = default;

  std::string get_key() const { return key_; }
  std::shared_ptr<M> get_value() const {
    if (IsAggregatedKVLayout(storage_layout)) {
      return value_;
    } else if (IsAggregatedKeyLayout(storage_layout)) {
      return map_ptr_->get_element(key_);
    } else {
      return nullptr;
    }
  }

 protected:
  // For AggregatedKVLayout
  IterItem(const std::string& key, const std::shared_ptr<M>& value)
      : key_(key), value_(value), map_ptr_(nullptr) {}

  // For AggregatedKeyLayout
  IterItem(const std::string& key, const std::shared_ptr<TMap>& map)
      : key_(key), value_(nullptr), map_ptr_(map) {}

 private:
  friend AlMapKeyIterator<TMap, M, storage_layout>;

  std::string key_;
  std::shared_ptr<M> value_;
  std::shared_ptr<TMap> map_ptr_;
};

template <class M, AlStorageLayout storage_layout>
class AlMapIterable final : public MyMap<M> {
 public:
  // No public constructors allowed. Use MyFactory to make instances instead.

  ~AlMapIterable() {}

  bool has_element(std::string element_key) override {
    // If meta data not exist, return false
    auto pmeta = get_meta_data(false /*revert_on_non_existence*/);
    if (!pmeta) {
      return false;
    }

    bool exists = false;
    if (IsAggregatedKVLayout(get_storage_layout())) {
      exists = (get_element_from_index(element_key) ==
                MapIterableErrors::MAP_ITERABLE_ERROR_NONE);
    } else if (IsAggregatedKeyLayout(get_storage_layout())) {
      exists = MyMap<M>::has_element(element_key);
    } else {
    }
    return exists;
  }

  std::shared_ptr<M> get_element(std::string element_key,
                                 bool revert_on_failure = true,
                                 bool only_in_memory
                                 [[gnu::unused]] = false) override {
    auto pmeta = get_meta_data(revert_on_failure);
    if (!pmeta) {
      return nullptr;
    }

    std::shared_ptr<MyBase> pvalue;
    int ret = MapIterableErrors::MAP_ITERABLE_ERROR_NONE;

    if (IsAggregatedKVLayout(storage_layout)) {
      std::shared_ptr<AggregatedValueM> pvalue_aggregated;
      ret = get_element_from_index(element_key, pvalue_aggregated);
      pvalue = pvalue_aggregated;
    } else if (IsAggregatedKeyLayout(storage_layout)) {
      pvalue = MyMap<M>::get_element(element_key, revert_on_failure);
    }

    if (revert_on_failure &&
        ret != MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
      std::string msg = "Failed to get element, key: " + element_key +
                        ", error code: " + std::to_string(ret);
      this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      this->mychain_lib_ptr_->MyRevert(msg);
    }

    return std::static_pointer_cast<M>(pvalue);
  }

  std::shared_ptr<M> add_element(std::string element_key,
                                 bool do_check
                                 [[gnu::unused]] = true) override {
    // If meta data not exist, create it
    auto pmeta = get_meta_data(false /*revert_on_non_existence*/);
    if (!pmeta) {
      create_meta_data();
    }

    if (has_element(element_key)) {
      delete_element(element_key);
    }

    std::shared_ptr<MyBase> pvalue;
    int ret = MapIterableErrors::MAP_ITERABLE_ERROR_NONE;

    if (IsAggregatedKVLayout(storage_layout)) {
      std::shared_ptr<AggregatedValueM> pvalue_aggregated;
      ret = add_element_key_to_index(element_key, pvalue_aggregated);
      pvalue = pvalue_aggregated;
    } else if (IsAggregatedKeyLayout(storage_layout)) {
      ret = add_element_key_to_index(element_key);
      pvalue = MyMap<M>::add_element(element_key, false /*do_check*/);
    }

    if (ret != MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
      std::string msg = "Error: failed to add element to index. element_key: " + element_key;
      this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      this->mychain_lib_ptr_->MyRevert(msg);
    }

    return std::static_pointer_cast<M>(pvalue);
  }

  int delete_element(std::string element_key,
                     bool revert_on_non_existence = true) override {
    // If meta data not exist, return early
    auto pmeta = get_meta_data(revert_on_non_existence);
    if (!pmeta) {
      return MapIterableErrors::MAP_ITERABLE_ERROR_META_DATA_NOT_EXISTS;
    }

    int ret = MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
    if (IsAggregatedKeyLayout(get_storage_layout())) {
      ret = MyMap<M>::delete_element(element_key, revert_on_non_existence);
      if (ret) {
        return ret;
      }
    }

    ret = delete_element_key_in_index(element_key);
    if (revert_on_non_existence &&
        ret == MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS) {
      std::string msg = "Error: element was deleted in map, but failed to be deleted in "
                        "index. element_key: " + element_key;
      this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      this->mychain_lib_ptr_->MyRevert(msg);
    }

    return ret;
  }

  size_t size() {
    size_t size = 0;

    // If meta data not exist, return 0;
    auto pmeta = get_meta_data(false /*revert_on_non_existence*/);
    if (!pmeta) {
      return 0;
    }

    for (uint16_t i = 0; i < get_index_width(); i++) {
      if (BitmapAccessor<AlMetaM>::read_bitmap(
              i, pmeta, get_bitmap_size(), "size", this->mychain_lib_ptr_)) {
        std::shared_ptr<AlIndexNodeM> pidx =
            get_index_node_by_index_key(std::to_string(i));
        if (pidx) {
          size += pidx->get_tcount();
        }
      }
    }
    return size;
  }

  AlMapKeyIterator<AlMapIterable<M, storage_layout>, M, storage_layout>
  get_map_key_iterator() {
    return AlMapKeyIterator<AlMapIterable<M, storage_layout>, M,
                            storage_layout>(
        std::static_pointer_cast<AlMapIterable<M, storage_layout>>(
            this->get_self_ptr()),
        this->mychain_lib_ptr_);
  }

  HashFunc get_hash_function() const {
    return meta_in_memory_.get_hash_function();
  }
  uint16_t get_index_width() const { return meta_in_memory_.get_index_width(); }
  uint16_t get_max_lcount() const { return meta_in_memory_.get_max_lcount(); }
  AlStorageLayout get_storage_layout() const {
    return meta_in_memory_.get_storage_layout();
  }

  uint16_t get_bitmap_size() const { return get_index_width() / 32 + 1; }

 protected:
  // Constructor for platform-layer and contract-layer mybuffer
  AlMapIterable(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
                std::unique_ptr<Key> _key, std::shared_ptr<MyBase> _parent,
                const HashFunc _hash_func, const uint16_t _index_width,
                const uint16_t _max_lcount, bool for_platform_layer,
                bool for_create_meta)
      : MyMap<M>(_mychain_lib_ptr, std::move(_key), _parent),
        meta_in_memory_(_hash_func, _index_width, _max_lcount, storage_layout,
                        _mychain_lib_ptr) {
    static_assert(std::is_base_of<MyBase, M>::value,
                  "Type M must derive from MyTable");

    const bool valid_for_contract_layer =
        IsAggregatedKeyLayout(storage_layout) &&
        (!std::is_same<AggregatedValueM, M>::value);
    if (!for_platform_layer && !valid_for_contract_layer) {
      std::string msg = "Only AggregatedKeyLayout is allowed in contract layer";
      _mychain_lib_ptr->MYLOG(LogLevel::kError, msg);
      _mychain_lib_ptr->MyRevert(msg);
    }

    const bool valid_for_platform_layer =
        (std::is_same<AggregatedValueM, M>::value ||
         std::is_same<SpreadedValueM, M>::value);
    if (for_platform_layer && !valid_for_platform_layer) {
      std::string msg = "Type M must be AggregatedValueM or SpreadedValueM";
      _mychain_lib_ptr->MYLOG(LogLevel::kError, msg);
      _mychain_lib_ptr->MyRevert(msg);
    }

    if (for_platform_layer) {
      // For the platform layer, we have to create or load the metadata here
      // before getting into the element granularity parallelism.
      if (for_create_meta) {
        create_meta_data();
      } else {
        get_meta_data();
      }
    } else {
      // For the contract layer, we delay the creation or loading of the
      // metadata to the map_iterable's APIs, since the parallel granularity of
      // the contract layer mybuffer is about root_tag or account, so it is
      // unnecessary to separate the `write` and `read` here when constructing
      // the map_iterable.
    }
  }

  std::shared_ptr<AlIndexNodeM> get_index_node_by_index_key(
      std::string _index_key,
      const std::shared_ptr<MyBase>& _parent = nullptr) {
    std::shared_ptr<AlIndexNodeM> rv;
    if (index_nodes_.find(_index_key) != index_nodes_.end()) {
      rv = index_nodes_[_index_key].lock();
    }
    if (!rv) {
      std::unique_ptr<Key> my_index_key =
          this->base_key_->get_map_index_key(_index_key);
      std::string key = my_index_key->to_string();
      std::string buf;
      int ret = this->mychain_lib_ptr_->MyGetStorage(key, buf);
      if (ret != 0) {
        return nullptr;
      } else {
        rv = MyFactory<AlIndexNodeM>::make_instance(
            this->mychain_lib_ptr_, buf, std::move(my_index_key),
            (_parent ? _parent : this->get_self_ptr()));
      }
      index_nodes_[_index_key] = rv;
    }
    return rv;
  }

  std::shared_ptr<AlIndexNodeM> add_index_node_by_index_key(
      std::string _index_key,
      const std::shared_ptr<MyBase>& _parent = nullptr) {
    std::shared_ptr<AlIndexNodeM> rv;
    std::unique_ptr<Key> my_index_key =
        this->base_key_->get_map_index_key(_index_key);
    rv = MyFactory<AlIndexNodeM>::make_instance(
        this->mychain_lib_ptr_, std::move(my_index_key),
        (_parent ? _parent : this->get_self_ptr()));
    BitmapAccessor<AlIndexNodeM>::fill_bitmap_with_zeros(
        get_bitmap_size(), rv, this->mychain_lib_ptr_);
    index_nodes_[_index_key] = rv;
    return rv;
  }

 private:
  friend MyFactory<AlMapIterable<M, storage_layout>>;
  friend AlMapKeyIterator<AlMapIterable<M, storage_layout>, M, storage_layout>;

  int get_naive_hash_value(const std::string& element_key,
                           const int level) const {
    int hash_val = 0;
    int l = level + 1;
    for (uint32_t i = 0; i < element_key.length(); i++) {
      hash_val += (uint8_t)element_key[i] * l;
    }
    hash_val %= get_index_width();
    return hash_val;
  }

  int get_city_hash_value(const std::string& element_key, int level) const {
    const std::string s = element_key + std::to_string(level);
    const uint32_t h = CityHash32(s.c_str(), s.size());
    return h % get_index_width();
  }

  int to_index_id(const std::string& element_key, const int level) const {
    if (get_hash_function() == HashFunc::HashFunc_city) {
      if (level == 0) {
        return get_naive_hash_value(element_key, level);
      } else {
        return get_city_hash_value(element_key, level);
      }
    } else if (get_hash_function() == HashFunc::HashFunc_city_all) {
      return get_city_hash_value(element_key, level);
    } else {
      std::string msg = "Unknow hash function: " + std::to_string(int(get_hash_function()));
      this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      this->mychain_lib_ptr_->MyRevert(msg);
      return -1;
    }
  }

  std::pair<int, std::string> get_index_id_and_key(
      const std::string& element_key, int level,
      const std::string& parent_index_key) {
    int index_id = to_index_id(element_key, level);
    if (parent_index_key.empty()) {
      return {index_id, std::to_string(index_id)};
    } else {
      return {index_id,
              Key::ConcatWithSlash(parent_index_key, std::to_string(index_id))};
    }
  }

  std::pair<int, std::string> get_parent_index_id_and_key(
      const std::string& element_key, int level,
      const std::string& child_index_key) {
    int index_id = to_index_id(element_key, level);
    int index_id_parent = to_index_id(element_key, level - 1);
    if (child_index_key.empty()) {
      return {index_id_parent, ""};
    } else {
      size_t pos = child_index_key.rfind(std::to_string(index_id));
      if (pos == std::string::npos || pos <= 1) {
        return {index_id_parent, ""};
      } else {
        return {index_id_parent, child_index_key.substr(0, pos - 1)};
      }
    }
  }

  int add_element_key_to_index(const std::string& element_key) {
    std::shared_ptr<AggregatedValueM> pvalue_not_used;
    return add_element_key_to_index(element_key, pvalue_not_used);
  }

  int add_element_key_to_index(const std::string& element_key,
                               std::shared_ptr<AggregatedValueM>& element) {
    std::shared_ptr<AlIndexNodeM> pindex;

    // Get meta.
    std::shared_ptr<AlMetaM> pmeta = get_meta_data();

    // Get the index_id and index_key of level 0.
    auto index_id_key =
        get_index_id_and_key(element_key, 0 /*level*/, "" /*parent_index_key*/);
    int index_id = index_id_key.first;
    std::string index_key = index_id_key.second;

    // If the bit at the expected position of the bitmap is 1,
    // get the index node.
    if (BitmapAccessor<AlMetaM>::read_bitmap(index_id, pmeta, get_bitmap_size(),
                                             "add_element",
                                             this->mychain_lib_ptr_)) {
      pindex = get_index_node_by_index_key(index_key);
    } else {
      // Else, set the bit to 1 and add the index node.
      BitmapAccessor<AlMetaM>::write_bitmap(index_id, true, pmeta,
                                            get_bitmap_size(), "add_element",

                                            this->mychain_lib_ptr_);
      pindex = add_index_node_by_index_key(index_key);
    }

    // Add the given element key to the index node of level 0.
    int ret = add_element_key_to_index_node(element_key, pindex, index_key,
                                            0 /*level*/, element);

    // Increase tcount in meta
    if (ret == MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
      pindex->set_tcount(pindex->get_tcount() + 1);
    }
    return ret;
  }

  int get_element_from_index(const std::string& element_key) {
    std::shared_ptr<AggregatedValueM> pvalue_not_used;
    return get_element_from_index(element_key, pvalue_not_used);
  }

  int get_element_from_index(const std::string& element_key,
                             std::shared_ptr<AggregatedValueM>& element) {
    int ret = MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
    std::shared_ptr<AlIndexNodeM> pindex;

    // Get meta.
    std::shared_ptr<AlMetaM> pmeta = get_meta_data();

    // Get the index_id and index_key of level 0.
    auto index_id_key =
        get_index_id_and_key(element_key, 0 /*level*/, "" /*parent_index_key*/);
    int index_id = index_id_key.first;
    std::string index_key = index_id_key.second;

    // If the bit at the expected position of the bitmap is 1,
    // get the index node.
    if (BitmapAccessor<AlMetaM>::read_bitmap(index_id, pmeta, get_bitmap_size(),
                                             "get_element",
                                             this->mychain_lib_ptr_)) {
      pindex = get_index_node_by_index_key(index_key);
      ret = get_element_from_index_node(element_key, pindex, index_key,
                                        0 /*level*/, element);
    } else {
      ret = MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS;
    }

    if (ret != MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
      element = nullptr;
    }
    return ret;
  }

  int delete_element_key_in_index(const std::string& element_key) {
    int ret = MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
    std::shared_ptr<AlIndexNodeM> pindex;

    // Get meta.
    std::shared_ptr<AlMetaM> pmeta = get_meta_data();

    // Get the index_id and index_key of level 0.
    auto index_id_key =
        get_index_id_and_key(element_key, 0 /*level*/, "" /*parent_index_key*/);
    int index_id = index_id_key.first;
    std::string index_key = index_id_key.second;

    // If the bit at the expected position of the bitmap is 1,
    // get the index node.
    if (BitmapAccessor<AlMetaM>::read_bitmap(index_id, pmeta, get_bitmap_size(),
                                             "delete_element",
                                             this->mychain_lib_ptr_)) {
      pindex = get_index_node_by_index_key(index_key);
      ret = delete_element_key_in_index_node(element_key, pindex, index_id,
                                             index_key, 0 /*level*/);
    } else {
      ret = MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS;
    }

    // Decrease tcount
    if (ret == MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
      if (!pindex->get_deleted()) {
        pindex->set_tcount(pindex->get_tcount() - 1);
      }
    }
    return ret;
  }

  int add_element_key_to_index_node(
      const std::string& element_key,
      const std::shared_ptr<AlIndexNodeM>& pindex, const std::string& index_key,
      const int level, std::shared_ptr<AggregatedValueM>& element) {
    // Get the size of the index node.
    uint16_t lcount = pindex->get_lcount();

    if (lcount < get_max_lcount()) {
      // If there is enough space for the element key, add it.
      auto pkey = pindex->get_keys()->append_element();
      pkey->set_val(element_key);

      if (IsAggregatedKVLayout(get_storage_layout())) {
        element = pindex->get_values()->append_element();
      }

      pindex->set_lcount(++lcount);
      return MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
    } else {
      // Else, check the bitmap and go to the next level.

      // Get the index_id and index_key of the next level.
      int level_child = level + 1;
      auto index_id_key = get_index_id_and_key(element_key, level_child,
                                               index_key /*parent_index_key*/);
      int index_id_child = index_id_key.first;
      std::string index_key_child = index_id_key.second;
      std::shared_ptr<AlIndexNodeM> pindex_child;

      // If the bit at the expected position of the bitmap is 1,
      // get the child index node.
      if (BitmapAccessor<AlIndexNodeM>::read_bitmap(
              index_id_child, pindex, get_bitmap_size(),
              "add_element_to_index_node", this->mychain_lib_ptr_)) {
        pindex_child = get_index_node_by_index_key(index_key_child, pindex);
      } else {
        // Else, set the bit to 1 and add the child index node.
        BitmapAccessor<AlIndexNodeM>::write_bitmap(
            index_id_child, true, pindex, get_bitmap_size(),
            "add_element_to_index_node", this->mychain_lib_ptr_);
        pindex_child = add_index_node_by_index_key(index_key_child, pindex);
      }

      // Add the element key to the index node of next level.
      return add_element_key_to_index_node(
          element_key, pindex_child, index_key_child, level_child, element);
    }
  }

  int delete_element_key_in_index_node(
      const std::string& element_key,
      const std::shared_ptr<AlIndexNodeM>& pindex, const uint16_t index_id,
      const std::string& index_key, const int level) {
    bool deleted = false;

    // Try to delete the element key in the index node's key list.
    int lcount = pindex->get_lcount();
    auto key_vector_ptr = pindex->get_keys();
    auto value_vector_ptr = pindex->get_values();

    // TODO: use more efficient way to delete the element key.
    for (int i = 0; i < lcount; i++) {
      auto pkey = key_vector_ptr->get_element(i);
      if (pkey->get_val() == element_key) {
        key_vector_ptr->delete_element(i);
        if (IsAggregatedKVLayout(get_storage_layout())) {
          value_vector_ptr->delete_element(i);
        }
        pindex->set_lcount(--lcount);
        deleted = true;
        break;
      }
    }

    if (deleted) {
      delete_empty_index_node_from_leaf(element_key, pindex, index_id,
                                        index_key, level);
      return MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
    } else {
      // Else, check the bitmap and go to the next level.

      // Get the index_id and index_key of the next level.
      int level_child = level + 1;
      auto index_id_key_child = get_index_id_and_key(
          element_key, level_child, index_key /*parent_index_key*/);
      int index_id_child = index_id_key_child.first;
      std::string index_key_child = index_id_key_child.second;
      std::shared_ptr<AlIndexNodeM> pindex_child;

      // If the bit at the expected position of the bitmap is 1, get the index
      // node.
      if (BitmapAccessor<AlIndexNodeM>::read_bitmap(
              index_id_child, pindex, get_bitmap_size(),
              "delete_element_to_index_node", this->mychain_lib_ptr_)) {
        pindex_child = get_index_node_by_index_key(index_key_child, pindex);
      } else {
        // else, the element key not exists, return false.
        return MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS;
      }

      // Delete the element key in the index node of next level.
      return delete_element_key_in_index_node(element_key, pindex_child,
                                              index_id_child, index_key_child,
                                              level_child);
    }
  }

  int get_element_from_index_node(const std::string& element_key,
                                  const std::shared_ptr<AlIndexNodeM>& pindex,
                                  const std::string& index_key, const int level,
                                  std::shared_ptr<AggregatedValueM>& element) {
    bool found = false;

    // Try to find the element key in the index node's key list.
    int lcount = pindex->get_lcount();
    auto key_vector_ptr = pindex->get_keys();
    auto value_vector_ptr = pindex->get_values();

    // TODO: use more efficient way to find the element key.
    for (int i = 0; i < lcount; i++) {
      auto pkey = key_vector_ptr->get_element(i);
      if (pkey->get_val() == element_key) {
        if (IsAggregatedKVLayout(get_storage_layout())) {
          element = value_vector_ptr->get_element(i);
        }
        found = true;
        break;
      }
    }

    if (found) {
      return MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
    } else {
      // Else, check the bitmap and go to the next level.

      // Get the index_id and index_key of the next level.
      int level_child = level + 1;
      auto index_id_key_child = get_index_id_and_key(
          element_key, level_child, index_key /*parent_index_key*/);
      int index_id_child = index_id_key_child.first;
      std::string index_key_child = index_id_key_child.second;
      std::shared_ptr<AlIndexNodeM> pindex_child;

      // If the bit at the expected position of the bitmap is 1, get the index
      // node.
      if (BitmapAccessor<AlIndexNodeM>::read_bitmap(
              index_id_child, pindex, get_bitmap_size(),
              "get_element_to_index_node", this->mychain_lib_ptr_)) {
        pindex_child = get_index_node_by_index_key(index_key_child, pindex);
      } else {
        // else, the element key not exists, return false.
        return MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS;
      }

      // find the element key in the index node of next level.
      return get_element_from_index_node(element_key, pindex_child,
                                         index_key_child, level_child, element);
    }
  }

  void delete_empty_index_node_from_leaf(
      const std::string& element_key,
      const std::shared_ptr<AlIndexNodeM>& pindex, const uint16_t index_id,
      const std::string& index_key, const int level) {
    // Check if the index node's keys and bitmap are both empty.
    bool empty = true;

    if (pindex->get_lcount() == 0) {
      // If there exist one bit that is not zero, it is not empty.
      for (uint16_t i = 0; i < get_index_width(); i++) {
        if (BitmapAccessor<AlIndexNodeM>::read_bitmap(
                i, pindex, get_bitmap_size(), "delete_empty_index",
                this->mychain_lib_ptr_)) {
          empty = false;
          break;
        }
      }
    } else {
      empty = false;
    }

    if (empty) {
      if (level <= 0) {
        // If the level 0 is empty, delete it and mark in meta.

        // Delete the index node.
        pindex->set_deleted(true);

        std::shared_ptr<AlMetaM> pmeta = get_meta_data();
        BitmapAccessor<AlMetaM>::write_bitmap(
            index_id, false, pmeta, get_bitmap_size(),
            "delete_empty_index at level 0", this->mychain_lib_ptr_);
      } else {
        // Else, delete the index node and check its parent.

        // Delete the index node.
        pindex->set_deleted(true);

        // Get the index key of parent index node.
        auto index_id_and_key_parent =
            get_parent_index_id_and_key(element_key, level, index_key);
        int index_id_parent = index_id_and_key_parent.first;
        std::string index_key_parent = index_id_and_key_parent.second;
        if (index_key_parent.empty()) {
          std::string msg = "Failed to find parent index key, element_key: " + element_key +
                            " level: " + std::to_string(level) + " index key: " + index_key;
          this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
          this->mychain_lib_ptr_->MyRevert(msg);
        }

        // Get the parent index node.
        std::shared_ptr<AlIndexNodeM> pindex_parent =
            get_index_node_by_index_key(index_key_parent);

        // Set the corresponding bit to zero in parent node's bitmap.
        BitmapAccessor<AlIndexNodeM>::write_bitmap(
            index_id, false, pindex_parent, get_bitmap_size(),
            "delete_empty_index at level 0", this->mychain_lib_ptr_);

        // Recursively delete empty parent nodes.
        delete_empty_index_node_from_leaf(element_key, pindex_parent,
                                          index_id_parent, index_key_parent,
                                          level - 1 /*level_parent*/);
      }
    }
    return;
  }

 protected:
  int create_meta_data() {
    // Read meta data from storage
    std::unique_ptr<Key> my_meta_key =
        this->base_key_->get_aldaba_map_meta_key();
    std::string key = my_meta_key->to_string();
    std::string buf;

    int ret = this->mychain_lib_ptr_->MyGetStorage(key, buf);
    if (ret == 0) {
      // Meta data already exist
      std::string msg = "MapIterable Meta Data already exitst! Key: " +
                        my_meta_key->printable().to_string();
      this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
      this->mychain_lib_ptr_->MyRevert(msg);
    } else {
      // Meta data not exist

      // Add meta data
      std::shared_ptr<AlMetaM> rv = MyFactory<AlMetaM>::make_instance(
          this->mychain_lib_ptr_, std::move(my_meta_key),
          this->get_parent_ptr());  // Make meta node as sibling of
                                    // map_iterable, or it can't be created when
                                    // constructing map_iterable

      // Set configs
      rv->set_hash_function(get_hash_function());
      rv->set_index_width(get_index_width());
      rv->set_max_lcount(get_max_lcount());
      rv->set_storage_layout(get_storage_layout());
      BitmapAccessor<AlMetaM>::fill_bitmap_with_zeros(get_bitmap_size(), rv,
                                                      this->mychain_lib_ptr_);

      pmeta_ = rv;
    }
    return 0;
  }

  std::shared_ptr<AlMetaM> get_meta_data(bool revert_on_non_existence = true) {
    std::shared_ptr<AlMetaM> rv;
    rv = pmeta_;
    if (!rv) {
      // Read meta data from storage
      std::unique_ptr<Key> my_meta_key =
          this->base_key_->get_aldaba_map_meta_key();
      std::string key = my_meta_key->to_string();
      std::string buf;

      int ret = this->mychain_lib_ptr_->MyGetStorage(key, buf);
      if (ret == 0) {
        // Meta data does exist
        rv = MyFactory<AlMetaM>::make_instance(
            this->mychain_lib_ptr_, buf, std::move(my_meta_key),
            this->get_parent_ptr());  // Make meta node as sibling of
                                      // map_iterable, or it can't be created
                                      // when constructing map_iterable

        // Set meta data in memory to the values read from storage.
        meta_in_memory_.MutateToMetaFromStorage(rv);

        pmeta_ = rv;
      } else {
        // Meta data does not exist
        if (revert_on_non_existence) {
          std::string msg = "MapIterable Meta Data do not exist!"
                            "Please call create_map_iterable(for platform layer)"
                            "or add_element(for contract layer) first"
                            "Key: " +
                            my_meta_key->printable().to_string();
          this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
          this->mychain_lib_ptr_->MyRevert(msg);
        }
      }
    }
    return rv;
  }

 private:
  class MetaInMemory {
   public:
    MetaInMemory(const HashFunc _hash_func, const uint16_t _index_width,
                 const uint16_t _max_lcount,
                 const AlStorageLayout _storage_layout,
                 std::shared_ptr<MychainLib> _mychain_lib_ptr)
        : hash_func_(_hash_func),
          index_width_(_index_width),
          max_lcount_(_max_lcount),
          storage_layout_(_storage_layout),
          mychain_lib_ptr_(_mychain_lib_ptr) {}

    ~MetaInMemory() {}

    void MutateToMetaFromStorage(const std::shared_ptr<AlMetaM>& pmeta) const {
      hash_func_ = pmeta->get_hash_function();
      index_width_ = pmeta->get_index_width();
      max_lcount_ = pmeta->get_max_lcount();
      storage_layout_ = pmeta->get_storage_layout();
    }

    HashFunc get_hash_function() const { return hash_func_; }
    uint16_t get_index_width() const {
      if (index_width_ == 0) {
        // Revert if invalid index_width
        std::string msg = "Invalid index width: 0";
        mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
        mychain_lib_ptr_->MyRevert(msg);

        // To avoid the wrong error report in C++ pinpoint scan.
        return 1;
      }

      return index_width_;
    }
    uint16_t get_max_lcount() const { return max_lcount_; }
    AlStorageLayout get_storage_layout() const { return storage_layout_; }

   private:
    mutable HashFunc hash_func_;
    mutable uint16_t index_width_;
    mutable uint16_t max_lcount_;
    mutable AlStorageLayout storage_layout_;
    std::shared_ptr<MychainLib> mychain_lib_ptr_;
  };

  std::map<std::string, std::weak_ptr<AlIndexNodeM>> index_nodes_;
  const MetaInMemory meta_in_memory_;
  std::shared_ptr<AlMetaM> pmeta_;
};

#endif  // MYBUFFER_ALDABA_MAPITERABLE_H
