#ifndef MYBUFFER_ALDABA_PLATFORM_H
#define MYBUFFER_ALDABA_PLATFORM_H

#include "mybuffer/mybuffer_aldaba_mapiterable.h"
#include "mybuffer/mybuffer_aldaba_platform_generated.h"
#include "mybuffer/dummylib.h"

constexpr char kAldabaRootKeyPrefix[] = "AldabaRoot";

// MapIterable types for Aldaba platform-layer mybuffer:
// * MyMapIterableForKVLayout
// * MyMapIterableForKeyLayout
typedef AlMapIterable<AggregatedValueM,
                      AlStorageLayout::AlStorageLayout_AggregatedKVLayout>
    MyMapIterableForKVLayout;
typedef AlMapIterable<SpreadedValueM,
                      AlStorageLayout::AlStorageLayout_AggregatedKeyLayout>
    MyMapIterableForKeyLayout;

// MapIterable Iterator types for Aldaba platform-layer mybuffer:
// * MyMapIteratorForKVLayout
// * MyMapIteratorForKeyLayout
typedef AlMapKeyIterator<MyMapIterableForKVLayout, AggregatedValueM,
                         AlStorageLayout::AlStorageLayout_AggregatedKVLayout>
    MyMapIteratorForKVLayout;
typedef AlMapKeyIterator<MyMapIterableForKeyLayout, SpreadedValueM,
                         AlStorageLayout::AlStorageLayout_AggregatedKeyLayout>
    MyMapIteratorForKeyLayout;

// Map types for Aldaba platform-layer mybuffer:
// * MyMapForSpreadedLayout
typedef MyMap<SpreadedValueM> MyMapForSpreadedLayout;

struct AldabaRootM;
typedef flatbuffers::Offset<AldabaRoot> AldabaRootMOffset;
typedef std::shared_ptr<AldabaRootM> AldabaRootMPtr;

struct AKVLMapIterableM : public MyTable<AKVLMapIterableT, AKVLMapIterable, false, true/*OnlyInMemory*/> {
private:
  std::weak_ptr<MyMapIterableForKVLayout > map_iterable_;
  using MyTable<AKVLMapIterableT, AKVLMapIterable, false, true>::MyTable;
  friend class MyFactory<AKVLMapIterableM, AKVLMapIterableT>;
  friend class MyFactory<AKVLMapIterableM>;
public:
  ~AKVLMapIterableM() {}
  std::shared_ptr<MyMapIterableForKVLayout> create_map_iterable(
    const std::string& hash_func,
    const uint16_t index_width,
    const uint16_t max_lcount) {
    std::shared_ptr<MyMapIterableForKVLayout> rv;
    rv = map_iterable_.lock();
    if (!rv) {
      rv = MyFactory<MyMapIterableForKVLayout>::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map_iterable"/*field_name*/),
          hash_func, index_width, max_lcount, true/*for_platform_layer*/, true/*for_create_meta*/,
          get_self_ptr());
      map_iterable_ = rv;
    }
    return rv;
  }
  std::shared_ptr<MyMapIterableForKVLayout> get_map_iterable() {
    std::shared_ptr<MyMapIterableForKVLayout> rv;
    rv = map_iterable_.lock();
    if (!rv) {
      rv = MyFactory<MyMapIterableForKVLayout>::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map_iterable"/*field_name*/),
          kHashFunc, kMapIndexWidth, kMapIndexMaxLcount, true/*for_platform_layer*/, false/*for_create_meta*/,
          get_self_ptr());
      map_iterable_ = rv;
    }
    return rv;
  }
};

struct AKLMapIterableM : public MyTable<AKLMapIterableT, AKLMapIterable, false, true/*OnlyInMemory*/> {
private:
  std::weak_ptr<MyMapIterableForKeyLayout> map_iterable_;
  using MyTable<AKLMapIterableT, AKLMapIterable, false, true>::MyTable;
  friend class MyFactory<AKLMapIterableM, AKLMapIterableT>;
  friend class MyFactory<AKLMapIterableM>;
public:
  ~AKLMapIterableM() {}
  std::shared_ptr<MyMapIterableForKeyLayout> create_map_iterable(
    const std::string& hash_func,
    const uint16_t index_width,
    const uint16_t max_lcount) {
    std::shared_ptr<MyMapIterableForKeyLayout> rv;
    rv = map_iterable_.lock();
    if (!rv) {
      rv = MyFactory<MyMapIterableForKeyLayout>::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map_iterable"/*field_name*/),
          hash_func, index_width, max_lcount, true/*for_platform_layer*/, true/*for_create_meta*/,
          get_self_ptr());
      map_iterable_ = rv;
    }
    return rv;
  }
  std::shared_ptr<MyMapIterableForKeyLayout> get_map_iterable() {
    std::shared_ptr<MyMapIterableForKeyLayout> rv;
    rv = map_iterable_.lock();
    if (!rv) {
      rv = MyFactory<MyMapIterableForKeyLayout>::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map_iterable"/*field_name*/),
          kHashFunc, kMapIndexWidth, kMapIndexMaxLcount, true/*for_platform_layer*/, false/*for_create_meta*/,
          get_self_ptr());
      map_iterable_ = rv;
    }
    return rv;
  }
};

struct SLMapM : public MyTable<SLMapT, SLMap, false, true/*OnlyInMemory*/> {
private:
  std::weak_ptr<MyMap<SpreadedValueM> > map_;
  using MyTable<SLMapT, SLMap, false, true>::MyTable;
  friend class MyFactory<SLMapM, SLMapT>;
  friend class MyFactory<SLMapM>;
public:
  ~SLMapM() {}
  std::shared_ptr<MyMap<SpreadedValueM>> get_map() {
    std::shared_ptr<MyMap<SpreadedValueM>> rv;
    rv = map_.lock();
    if (!rv) {
      rv = MyFactory<MyMap<SpreadedValueM> >::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map"/*field_name*/), get_self_ptr());
      map_ = rv;
    }
    return rv;
  }
};

typedef flatbuffers::Offset<AKVLMapIterable> AKVLMapIterableMOffset;
typedef std::shared_ptr<AKVLMapIterableM> AKVLMapIterableMPtr;
typedef std::shared_ptr<MyMap<AKVLMapIterableM> > AKVLMapIterableMMapPtr;

typedef flatbuffers::Offset<AKLMapIterable> AKLMapIterableMOffset;
typedef std::shared_ptr<AKLMapIterableM> AKLMapIterableMPtr;
typedef std::shared_ptr<MyMap<AKLMapIterableM> > AKLMapIterableMMapPtr;

typedef flatbuffers::Offset<SLMap> SLMapMOffset;
typedef std::shared_ptr<SLMapM> SLMapMPtr;
typedef std::shared_ptr<MyMap<SLMapM> > SLMapMMapPtr;

struct AldabaRootM : public MyTable<AldabaRootT, AldabaRoot, false, true/*OnlyInMemory*/> {
private:
  std::weak_ptr<MyMap<AKVLMapIterableM> > aggregated_kv_layout_map_iterables_;
  std::weak_ptr<MyMap<AKLMapIterableM> > aggregated_key_layout_map_iterables_;
  std::weak_ptr<MyMap<SLMapM> > spreaded_maps_;
  using MyTable<AldabaRootT, AldabaRoot, false, true>::MyTable;
  friend class MyFactory<AldabaRootM, AldabaRootT>;
  friend class MyFactory<AldabaRootM>;
public:
  ~AldabaRootM() {}

  template<class MapIterableType>
  std::shared_ptr<MapIterableType> create_map_iterable(
    const std::string& name,
    const std::string& hash_func = kHashFunc,
    const uint16_t index_width = kMapIndexWidth,
    const uint16_t max_lcount = kMapIndexMaxLcount);

  template<class MapIterableType>
  std::shared_ptr<MapIterableType> get_map_iterable(const std::string& name);

  template<class MapType>
  std::shared_ptr<MapType> get_map(const std::string& name);

  template<class MapIterableType>
  static std::string get_map_iterable_meta_key(const std::string& map_iterable_name);

  template<class MapType>
  static std::string get_map_prefetch_key(const std::string& map_name, const std::string& element_key);

private:
  // NOTE: the following functions have the field id's hard-coded and assume
  // that the later updates to mybuffer_aldaba_platform.fbs will
  // follow the rules of schema compatibility
  static std::string get_map_iterable_meta_key_from(int field_id, const std::string& field_name, const std::string& map_iterable_name) {
    std::unique_ptr<Key> root_key = Key::get_root_table_key(kAldabaRootKeyPrefix, std::make_shared<DummyLib>());
    return root_key
    ->get_table_field_key(FieldNodeType::MAP_V2, field_id, field_name)
    ->get_map_element_key(map_iterable_name)
    ->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map_iterable"/*field_name*/)
    ->get_aldaba_map_meta_key()
    ->to_string();
  }

  static std::string get_map_prefetch_key_from(int field_id, const std::string& field_name, const std::string& map_name, const std::string& element_key) {
    std::unique_ptr<Key> root_key = Key::get_root_table_key(kAldabaRootKeyPrefix, std::make_shared<DummyLib>());
    return root_key
    ->get_table_field_key(FieldNodeType::MAP_V2, field_id, field_name)
    ->get_map_element_key(map_name)
    ->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "map"/*field_name*/)
    ->get_map_element_key(element_key)
    ->to_string();
  }

  AKVLMapIterableMMapPtr get_aggregated_kv_map_iterables() {
    AKVLMapIterableMMapPtr rv;
    rv = aggregated_kv_layout_map_iterables_.lock();
    if (!rv) {
      rv = MyFactory<MyMap<AKVLMapIterableM> >::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 0/*field_id*/, "aggregated_kv_map_iterables"/*field_name*/), get_self_ptr());
      aggregated_kv_layout_map_iterables_ = rv;
    }
    return rv;
  }
  AKLMapIterableMMapPtr get_aggregated_key_map_iterables() {
    AKLMapIterableMMapPtr rv;
    rv = aggregated_key_layout_map_iterables_.lock();
    if (!rv) {
      rv = MyFactory<MyMap<AKLMapIterableM> >::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 1/*field_id*/, "aggregated_key_map_iterables"/*field_name*/), get_self_ptr());
      aggregated_key_layout_map_iterables_ = rv;
    }
    return rv;
  }
  SLMapMMapPtr get_spreaded_maps() {
    SLMapMMapPtr rv;
    rv = spreaded_maps_.lock();
    if (!rv) {
      rv = MyFactory<MyMap<SLMapM> >::make_instance(mychain_lib_ptr_,
          base_key_->get_table_field_key(FieldNodeType::MAP_V2, 2/*field_id*/, "spreaded_maps"/*field_name*/), get_self_ptr());
      spreaded_maps_ = rv;
    }
    return rv;
  }
};

template<>
inline std::shared_ptr<MyMapIterableForKVLayout> AldabaRootM::create_map_iterable<MyMapIterableForKVLayout>(
  const std::string& name,
  const std::string& hash_func,
  const uint16_t index_width,
  const uint16_t max_lcount) {
  AKVLMapIterableMPtr instance = get_aggregated_kv_map_iterables()->get_element(name, false/*revert_on_failure*/, true/*only_in_memory*/);
  if (!instance) {
    // Since AKVLMapIterableM is OnlyInMemory, add_element() won't cause any write I/O.
    instance = get_aggregated_kv_map_iterables()->add_element(name);
  }
  return instance->create_map_iterable(hash_func, index_width, max_lcount);
}

template<>
inline std::shared_ptr<MyMapIterableForKVLayout> AldabaRootM::get_map_iterable<MyMapIterableForKVLayout>(
  const std::string& name) {
  AKVLMapIterableMPtr instance = get_aggregated_kv_map_iterables()->get_element(name, false/*revert_on_failure*/, true/*only_in_memory*/);
  if (!instance) {
    // Since AKVLMapIterableM is OnlyInMemory, add_element() won't cause any write I/O.
    instance = get_aggregated_kv_map_iterables()->add_element(name);
  }
  return instance->get_map_iterable();
}

template<>
inline std::shared_ptr<MyMapIterableForKeyLayout> AldabaRootM::create_map_iterable<MyMapIterableForKeyLayout>(
  const std::string& name,
  const std::string& hash_func,
  const uint16_t index_width,
  const uint16_t max_lcount) {
  AKLMapIterableMPtr instance = get_aggregated_key_map_iterables()->get_element(name, false/*revert_on_failure*/, true/*only_in_memory*/);
  if (!instance) {
    // Since AKLMapIterableM is OnlyInMemory, add_element() won't cause any write I/O.
    instance = get_aggregated_key_map_iterables()->add_element(name);
  }
  return instance->create_map_iterable(hash_func, index_width, max_lcount);
}

template<>
inline std::shared_ptr<MyMapIterableForKeyLayout> AldabaRootM::get_map_iterable<MyMapIterableForKeyLayout>(
  const std::string& name) {
  AKLMapIterableMPtr instance = get_aggregated_key_map_iterables()->get_element(name, false/*revert_on_failure*/, true/*only_in_memory*/);
  if (!instance) {
    // Since AKLMapIterableM is OnlyInMemory, add_element() won't cause any write I/O.
    instance = get_aggregated_key_map_iterables()->add_element(name);
  }
  return instance->get_map_iterable();
}

template<>
inline std::shared_ptr<MyMapForSpreadedLayout> AldabaRootM::get_map<MyMapForSpreadedLayout>(const std::string& name) {
  SLMapMPtr instance = get_spreaded_maps()->get_element(name, false/*revert_on_failure*/, true/*only_in_memory*/);
  if (!instance) {
    // Since SLMapM is OnlyInMemory, add_element() won't cause any write I/O.
    instance = get_spreaded_maps()->add_element(name);
  }
  return instance->get_map();
}

template<>
inline std::string AldabaRootM::get_map_iterable_meta_key<MyMapIterableForKVLayout>(const std::string& map_iterable_name) {
  return get_map_iterable_meta_key_from(0/*field_id*/, "aggregated_kv_map_iterables"/*field_name*/, map_iterable_name);
}

template<>
inline std::string AldabaRootM::get_map_iterable_meta_key<MyMapIterableForKeyLayout>(const std::string& map_iterable_name) {
  return get_map_iterable_meta_key_from(1/*field_id*/, "aggregated_key_map_iterables"/*field_name*/, map_iterable_name);
}

template<>
inline std::string AldabaRootM::get_map_prefetch_key<MyMapForSpreadedLayout>(const std::string& map_name, const std::string& element_key) {
  return get_map_prefetch_key_from(2/*field_id*/, "spreaded_maps"/*field_name*/, map_name, element_key);
}

class AldabaRootManager {
public:
  AldabaRootManager() = default;
  AldabaRootManager(const AldabaRootManager& ) = delete;
  AldabaRootManager& operator=(const AldabaRootManager&) = delete;

  ~AldabaRootManager() {
    for (const auto& kv : roots_) {
      std::shared_ptr<AldabaRootM> root = kv.second.lock();
      if (root) {
        root->clean_all();
      }
    }
  }

  AldabaRootMPtr GetAldabaRootM(const std::string& prefix, std::shared_ptr<MychainLib> mychain_lib_ptr);

private:
  std::map<std::string, std::weak_ptr<AldabaRootM>> roots_;
};

inline AldabaRootMPtr AldabaRootManager::GetAldabaRootM(const std::string& prefix, std::shared_ptr<MychainLib> mychain_lib_ptr) {
  AldabaRootMPtr sp;
  std::weak_ptr<AldabaRootM> wp;

  if (roots_.find(prefix) != roots_.end()) {
    wp = roots_[prefix];
    sp = wp.lock();
  }

  if (!sp) {
    mychain_lib_ptr->SetAllowDotInMapKey(false);
    mychain_lib_ptr->SetManualCommit(true);
    std::unique_ptr<Key> root_key = Key::get_root_table_key(kAldabaRootKeyPrefix, mychain_lib_ptr);
    sp = MyFactory<AldabaRootM>::make_instance(mychain_lib_ptr, std::move(root_key) /*, nullptr*/);
    wp = sp;
    roots_[prefix] = wp;
  }
  return sp;
}

#endif // MYBUFFER_ALDABA_PLATFORM_H
