#ifndef MYBUFFER_MYFACTORY_H
#define MYBUFFER_MYFACTORY_H

#include "flatbuffers/flatbuffers.h"
#include "mybuffer/mybuffer_const_define.h"
#include "mybuffer/mybuffer_hashfunc_generated.h"
#include "mybuffer/mybuffer_aldaba_storagelayout_generated.h"
#include "mybuffer/mybuffer_mybase.h"
#include "mybuffer/mychainlib.h"

#include <memory>

enum MapIterableErrors {
    MAP_ITERABLE_ERROR_NONE = 0,
    MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS = 1,
    MAP_ITERABLE_ERROR_META_DATA_NOT_EXISTS = 2,
};

inline HashFunc HashFuncFromString(const std::string& hash_func_name) {
    if (hash_func_name == "city") {
        return HashFunc::HashFunc_city;
    } else if (hash_func_name == "city_all") {
        return HashFunc::HashFunc_city_all;
    } else {
        return HashFunc::HashFunc_invalid;
    }
}

template <class M, class TNative = flatbuffers::NativeTable>
class MyFactory {
public:
    // For brand-new serialization tree roots or MyMap nodes
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, const std::shared_ptr<MyBase>& _parent=nullptr) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance MyMap node: key %s", _key->to_string().c_str());
        // Setting order: is_root_ -> self_ptr_/root_ptr_ -> dirty_
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, std::move(_key), _parent));
        rv->set_self_ptr(rv);
        rv->set_dirty(true);
        return rv;
    }

    // For MyMapIterable
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, const std::string& hash_func_name, const uint16_t _index_width, const uint16_t _max_lcount,
        const std::shared_ptr<MyBase>& _parent=nullptr) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance MyMapIterable node: key %s func_name %s index_width %d max_lcount %d",
            _key->to_string().c_str(), hash_func_name.c_str(), _index_width, _max_lcount);
        HashFunc _hash_func = HashFuncFromString(hash_func_name);
        // Check the validity of configs
        if ((_hash_func == HashFunc::HashFunc_invalid) ||
            (_index_width == 0) ||
            (_max_lcount == 0)) {
            const std::string msg = "Invalid configs: "
                " hash_func_name given: " + hash_func_name +
                " index_width given: " + std::to_string(_index_width) +
                " max_lcount given: " + std::to_string(_max_lcount);
            _mychain_lib_ptr->MYLOG(LogLevel::kError, msg);
            _mychain_lib_ptr->MyRevert(msg);
        }
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, std::move(_key), _parent, _hash_func, _index_width, _max_lcount));
        rv->set_self_ptr(rv);
        rv->set_dirty(true);
        return rv;
    }

    // For AlMapIterable nodes
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, const std::string& hash_func_name, const uint16_t _index_width, const uint16_t _max_lcount,
        const bool for_platform_layer, const bool for_create_meta,
        const std::shared_ptr<MyBase>& _parent=nullptr) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance AlMapIterable node: key %s func_name %s index_width %d max_lcount %d",
            _key->to_string().c_str(), hash_func_name.c_str(), _index_width, _max_lcount);
        HashFunc _hash_func = HashFuncFromString(hash_func_name);
        // Check the validity of configs
        if ((_hash_func == HashFunc::HashFunc_invalid) ||
            (_index_width == 0) ||
            (_max_lcount == 0)) {
            const std::string msg = "Invalid configs: "
                " hash_func_name given: " + hash_func_name +
                " index_width given: " + std::to_string(_index_width) +
                " max_lcount given: " + std::to_string(_max_lcount);
            _mychain_lib_ptr->MYLOG(LogLevel::kError, msg);
            _mychain_lib_ptr->MyRevert(msg);
        }
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, std::move(_key), _parent, _hash_func, _index_width, _max_lcount, for_platform_layer, for_create_meta));
        rv->set_self_ptr(rv);
        rv->set_dirty(true);
        return rv;
    }

    // For existing (on-disk) serialization tree roots
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        const std::string& _buf, std::unique_ptr<Key> _key, const std::shared_ptr<MyBase>& _parent=nullptr) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance root node: key %s", _key->to_string().c_str());
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, _buf, std::move(_key), _parent));
        rv->set_self_ptr(rv);
        return rv;
    }

    // For non-root nodes
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, std::unique_ptr<TNative>& _upnative, std::shared_ptr<MyBase> _parent) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance non-root node: key %s", _key->to_string().c_str());
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, std::move(_key), _upnative, _parent));
        rv->set_self_ptr(rv);
        return rv;
    }

    // For MyVector
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, std::vector<std::unique_ptr<TNative> >& _native_elements,
        std::shared_ptr<MyBase> _parent, uint32_t _size_limit = kDefaultVectorSizeLimit) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance MyVector(non-primitive): key %s", _key->to_string().c_str());
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, std::move(_key), _native_elements, _parent, _size_limit));
        rv->set_self_ptr(rv);
        return rv;
    }

    // For MyVector of Primitive types
    static std::shared_ptr<M> make_instance(const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, std::vector<TNative>& _native_elements,
        std::shared_ptr<MyBase> _parent, uint32_t _size_limit = kDefaultVectorSizeLimit) {
        _mychain_lib_ptr->MYLOG(LogLevel::kDebug, "MyFactory make_instance MyVector(primitive)): key %s", _key->to_string().c_str());
        std::shared_ptr<M> rv(new M(_mychain_lib_ptr, std::move(_key), _native_elements, _parent, _size_limit));
        rv->set_self_ptr(rv);
        return rv;
    }

};

#endif // MYBUFFER_MYFACTORY_H
