#ifndef MYBUFFER_MYMAPITERABLE_H
#define MYBUFFER_MYMAPITERABLE_H

#include "flatbuffers/flatbuffers.h"
#include "mybuffer/key.h"
#include "mybuffer/mybuffer_const_define.h"
#include "mybuffer/mybuffer_metadata_generated.h"
#include "mybuffer/mybuffer_mybase.h"
#include "mybuffer/mybuffer_myfactory.h"
#include "mybuffer/mybuffer_mytable.h"
#include "mybuffer/mybuffer_myvector.h"
#include "mybuffer/mybuffer_mymap.h"
#include "mybuffer/mychainlib.h"

#include <third_party/cityhash/city.h>

#include <map>
#include <memory>
#include <deque>

#ifndef MYBUFFER_MAPITERABLE_FORCE_UPDATE_META_BITMAP
#define MYBUFFER_MAPITERABLE_FORCE_UPDATE_META_BITMAP false
#endif


typedef std::shared_ptr<MyVector<uint32_t, uint32_t, true>> MetaBitmapPtr;
class MyBitmapAccessor {
public:
    static void init_bitmap(uint16_t index_width, MetaBitmapPtr bitmap) {
        bitmap->clear();
        for (uint32_t i = 0; i < bitmap_size_of_index_width(index_width); i++) {
            bitmap->append_element(0);
        }
    }

    static bool check_bitmap_size(uint16_t index_width, MetaBitmapPtr bitmap) {
        return bitmap->size() == bitmap_size_of_index_width(index_width);
    }

    static bool get_bit(uint16_t index, MetaBitmapPtr bitmap) {
        uint16_t move = (31 - index%32);
        uint16_t pos  = index/32;
        uint32_t bitmask = 1 << move;
        return (bool)((bitmap->get_element(pos) & bitmask) >> (move));
    }

    static void set_bit(uint16_t index, bool bit, MetaBitmapPtr bitmap) {
        uint16_t move = (31 - index%32);
        uint16_t pos  = index/32;
        uint32_t bitmask = bit ? (1 << move) : (~(1 << move));
        uint32_t val = bitmap->get_element(pos);
        bitmap->set_element(pos, bit ? (val|bitmask) : (val&bitmask));
        return;
    }

    template<typename TMap>
    static void update_bitmap(TMap* map, uint16_t index_width, MetaBitmapPtr bitmap) {
        bool valid_bitmap_size = MyBitmapAccessor::check_bitmap_size(index_width, bitmap);
        bool force_update = MYBUFFER_MAPITERABLE_FORCE_UPDATE_META_BITMAP;
        if (!valid_bitmap_size || force_update) {
            MyBitmapAccessor::init_bitmap(index_width, bitmap);

            // update bitmap by scanning nodes of level 0
            for (uint16_t i = 0; i < index_width; i++) {
                auto pindex = map->get_index_node_by_index_key(std::to_string(i));
                if (pindex && ((pindex->get_tcount() > 0) || (pindex->get_lcount() > 0))) {
                    set_bit(i, true, bitmap);
                }
            }
        }
    }

    static void update_bitmap(uint16_t index_width, const std::vector<bool>& src, MetaBitmapPtr dst) {
        bool valid_bitmap_size = MyBitmapAccessor::check_bitmap_size(index_width, dst);
        bool force_update = MYBUFFER_MAPITERABLE_FORCE_UPDATE_META_BITMAP;
        if (!valid_bitmap_size || force_update) {
            MyBitmapAccessor::init_bitmap(index_width, dst);
            if (index_width == src.size()) {
                for (uint16_t i = 0; i < index_width; i++) {
                    MyBitmapAccessor::set_bit(i, src[i], dst);
                }
            }
        }
    }

private:
    static uint32_t bitmap_size_of_index_width(uint16_t index_width) {
        return index_width/32+1;
    }
};

class MyIndexNodeStringM : public MyTable<MyIndexNodeStringT, MyIndexNodeString> {
public:
    // No public constructors allowed. Use MyFactory to make instances instead.

    ~MyIndexNodeStringM() {}

    std::string get_val() {
        return native_table_ptr_->val;
    }

    bool set_val(std::string _val) {
        if (_val.length() > kDefaultStringSizeLimit) {
            std::string msg = "The string size is over the limit" + std::to_string(kDefaultStringSizeLimit);
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        set_dirty(true);
        native_table_ptr_->val = std::move(_val);
        return true;
    }

protected:
    using MyTable<MyIndexNodeStringT, MyIndexNodeString>::MyTable;

private:
    template <class M, class T>
    friend class MyFactory;
};


class MyIndexNodeIntM : public MyTable<MyIndexNodeIntT, MyIndexNodeInt> {
public:
    // No public constructors allowed. Use MyFactory to make instances instead.

    ~MyIndexNodeIntM() {}

    int32_t get_val() {
        return native_table_ptr_->val;
    }

    bool set_val(int32_t _val) {
        set_dirty(true);
        native_table_ptr_->val = _val;
        return true;
    }

protected:
    using MyTable<MyIndexNodeIntT, MyIndexNodeInt>::MyTable;

private:
    template <class M, class T>
    friend class MyFactory;
};


class MyIndexNodeM : public MyTable<MyIndexNodeT, MyIndexNode> {
public:
    // No public constructors allowed. Use MyFactory to make instances instead.

    ~MyIndexNodeM() {}

    std::shared_ptr<MyVector<MyIndexNodeStringM, MyIndexNodeStringT> > get_keys() {
        std::shared_ptr<MyVector<MyIndexNodeStringM, MyIndexNodeStringT> > rv;
        rv = keys_.lock();
        if (!rv) {
            rv = MyFactory<MyVector<MyIndexNodeStringM, MyIndexNodeStringT>, MyIndexNodeStringT>::make_instance(
                mychain_lib_ptr_,base_key_->get_table_field_key(FieldNodeType::VECTOR, 0),
                native_table_ptr_->keys, get_self_ptr());
            keys_ = rv;
        }
        return rv;
    }

    // `version` and `width` is not in use.
    // uint16_t get_version() {
    //     return native_table_ptr_->version;
    // }

    // bool set_version(uint16_t _version) {
    //     set_dirty(true);
    //     native_table_ptr_->version = _version;
    //     return true;
    // }

    // uint16_t get_width() {
    //     return native_table_ptr_->width;
    // }

    // bool set_width(uint16_t _width) {
    //     set_dirty(true);
    //     native_table_ptr_->width = _width;
    //     return true;
    // }

    uint16_t get_lcount() {
        return native_table_ptr_->lcount;
    }

    bool set_lcount(uint16_t _lcount) {
        set_dirty(true);
        native_table_ptr_->lcount = _lcount;
        return true;
    }

    uint64_t get_tcount() {
        return native_table_ptr_->tcount;
    }

    bool set_tcount(uint64_t _tcount) {
        set_dirty(true);
        native_table_ptr_->tcount = _tcount;
        return true;
    }

    std::shared_ptr<MyVector<MyIndexNodeIntM, MyIndexNodeIntT> > get_bitmap() {
        std::shared_ptr<MyVector<MyIndexNodeIntM, MyIndexNodeIntT> > rv;
        rv = bitmap_.lock();
        if (!rv) {
            rv = MyFactory<MyVector<MyIndexNodeIntM, MyIndexNodeIntT>, MyIndexNodeIntT>::make_instance(
                mychain_lib_ptr_,base_key_->get_table_field_key(FieldNodeType::VECTOR, 4),
                native_table_ptr_->bitmap, get_self_ptr());
            bitmap_ = rv;
        }
        return rv;
    }

    /**
     * @brief index_width_to_bitmap_size returns the corresponding bitmap size
     * of given index_width.
     *
     * @param index_width
     * @return bitmap_size
    */
    static uint16_t index_width_to_bitmap_size(uint16_t index_width) {
        return index_width/32 + 1;
    }

    /**
     * @brief get_bit_by_index_noed_id returns the corresponding bit of the
     * index_node_id in bitmap.
     *
     * @param index_node_id
     * @param bitmap
     *
     * @return boolean, true for 1, false for 0.
    */
    bool get_bit_by_index_node_id(
        uint16_t index_node_id,
        std::shared_ptr<MyVector<MyIndexNodeIntM, MyIndexNodeIntT>> bitmap) const {
        uint16_t move = (31 - index_node_id%32);
        uint16_t pos  = index_node_id/32;
        uint32_t bitmask = 1 << move;
        return (bool)((bitmap->get_element(pos)->get_val() & bitmask) >> (move));
    }

    /**
     * @brief set_bit_by_index_noed_id sets the corresponding bit of the
     * index_node_id in bitmap.
     *
     * @param index_node_id
     * @param bit true for 1, false for 0.
     *
    */
    void set_bit_by_index_node_id(
        uint16_t index_node_id,
        bool bit,
        std::shared_ptr<MyVector<MyIndexNodeIntM, MyIndexNodeIntT>> bitmap) const {
        uint16_t move = (31 - index_node_id%32);
        uint16_t pos  = index_node_id/32;
        uint32_t bitmask = bit ? (1 << move) : (~(1 << move));
        int32_t val = bitmap->get_element(pos)->get_val();
        bitmap->get_element(pos)->set_val(bit ? (val|bitmask) : (val&bitmask));
        return;
    }

protected:
    using MyTable<MyIndexNodeT, MyIndexNode>::MyTable;

private:
    template <class M, class T>
    friend class MyFactory;

private:
    std::weak_ptr<MyVector<MyIndexNodeStringM, MyIndexNodeStringT> > keys_;
    std::weak_ptr<MyVector<MyIndexNodeIntM, MyIndexNodeIntT> > bitmap_;
};

struct MyMapIterableMetaM : public MyTable<MyMapIterableMetaT, MyMapIterableMeta> {
private:
  std::weak_ptr<MyVector<uint32_t, uint32_t, true> > bitmap_;
  using MyTable<MyMapIterableMetaT, MyMapIterableMeta>::MyTable;
  friend class MyFactory<MyMapIterableMetaM, MyMapIterableMetaT>;
  friend class MyFactory<MyMapIterableMetaM>;
public:
  ~MyMapIterableMetaM() {}
  HashFunc get_hash_function() {
    return native_table_ptr_->hash_function;
  }
  bool set_hash_function(HashFunc _hash_function) {
    set_dirty(true);
    native_table_ptr_->hash_function = _hash_function;
    return true;
  }
  uint16_t get_index_width() {
    return native_table_ptr_->index_width;
  }
  bool set_index_width(uint16_t _index_width) {
    set_dirty(true);
    native_table_ptr_->index_width = _index_width;
    return true;
  }
  uint16_t get_max_lcount() {
    return native_table_ptr_->max_lcount;
  }
  bool set_max_lcount(uint16_t _max_lcount) {
    set_dirty(true);
    native_table_ptr_->max_lcount = _max_lcount;
    return true;
  }
  std::shared_ptr<MyVector<uint32_t, uint32_t, true>> get_bitmap() {
    std::shared_ptr<MyVector<uint32_t, uint32_t, true>> rv;
    rv = bitmap_.lock();
    if (!rv) {
      rv = MyFactory<MyVector<uint32_t, uint32_t, true>, uint32_t>::make_instance(
           mychain_lib_ptr_, base_key_->get_table_field_key(FieldNodeType::VECTOR, 3/*field_id*/),
           native_table_ptr_->bitmap, get_self_ptr(), 0);
      bitmap_ = rv;
    }
    return rv;
  }
};


template<class TMap>
class MapKeyIterator {
public:
    MapKeyIterator(const std::shared_ptr<TMap>& _map_ptr, const std::shared_ptr<MychainLib>& _mychain_lib_ptr)
    : map_ptr_(_map_ptr), mychain_lib_ptr_(_mychain_lib_ptr) {
        _map_ptr->update_meta_data(
            _map_ptr->get_hash_function(), _map_ptr->get_index_width(), _map_ptr->get_max_lcount(), true/*update_bitmap*/);
        auto pmeta = _map_ptr->get_meta_data();
        for (uint16_t i=0; i<_map_ptr->get_index_width(); i++) {
            // Metadata Bitmap: access index nodes of level 0 according to the bitmap of metadata.
            bool exists = MyBitmapAccessor::get_bit(i, pmeta->get_bitmap());
            if (!exists) {
                continue;
            }

            std::shared_ptr<MyIndexNodeM> pindex = map_ptr_->get_index_node_by_index_key(std::to_string(i));
            if (pindex) {
                nodes_.push_back(std::unique_ptr<Node>(new Node(pindex)));
            }
        }
        ++(*this);
    }

    ~MapKeyIterator() {}

    bool valid() {
        std::shared_ptr<MyIndexNodeM> pindex;
        return isValid_ && pnode_ && (pindex = pnode_->index_node_ptr_) && (offset_ < pindex->get_lcount());
    }

    std::string operator*() {
        if (!valid()) {
            std::string msg = "The iterator is invalid.";
            mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            mychain_lib_ptr_->MyRevert(msg);
        }
        std::shared_ptr<MyIndexNodeM> pindex = pnode_->index_node_ptr_;
        std::shared_ptr<MyIndexNodeStringM> pkey = pindex->get_keys()->get_element(offset_);
        return pkey->get_val();
    }

    MapKeyIterator<TMap>& operator++() {
        std::shared_ptr<MyIndexNodeM> pindex;
        isValid_ = false;
        if (pnode_ && (pindex = pnode_->index_node_ptr_) && (offset_ < pindex->get_lcount() - 1)) {
            offset_++;
            isValid_ = true;
        } else {
            while (!nodes_.empty()) {
                pnode_ = std::move(nodes_.front());
                nodes_.pop_front();

                // Enqueue children tree nodes
                if ((pindex = pnode_->index_node_ptr_)) {

                    // If the size of the bitmap is not valid, revert
                    uint16_t bitmap_size = pindex->index_width_to_bitmap_size(map_ptr_->get_index_width());
                    if (pindex->get_bitmap()->size() != bitmap_size) {
                        std::string msg = "Invalid bitmap size when iterator++."
                                          "got: "+std::to_string(pindex->get_bitmap()->size())+
                                          " want: "+std::to_string(bitmap_size);
                        this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                        this->mychain_lib_ptr_->MyRevert(msg);
                    }

                    for (uint16_t i = 0; i<map_ptr_->get_index_width(); i++) {
                        if (pindex->get_bit_by_index_node_id(i, pindex->get_bitmap())) {
                            std::string index_key = pindex->get_base_key().substr(map_ptr_->get_base_key().size()+1);
                            std::shared_ptr<MyIndexNodeM> pidx =
                                map_ptr_->get_index_node_by_index_key(Key::ConcatWithSlash(index_key, std::to_string(i)));
                            if (pidx) {
                                nodes_.push_back(std::unique_ptr<Node>(new Node(pidx)));
                            }
                        }
                    }
                }

                if ((pindex = pnode_->index_node_ptr_) && pindex->get_lcount() > 0) {
                    offset_ = 0;
                    isValid_ = true;
                    break;
                }
            }
        }
        return *this;
    }

private:
    struct Node {
        std::shared_ptr<MyIndexNodeM> index_node_ptr_;

        explicit Node(std::shared_ptr<MyIndexNodeM> _ptr) {
            index_node_ptr_ = _ptr;
        }

        ~Node() {
            if (index_node_ptr_)
                index_node_ptr_.reset();
        }
    };

    std::shared_ptr<TMap> map_ptr_;
    std::deque<std::unique_ptr<Node> > nodes_;
    std::unique_ptr<Node> pnode_;
    int offset_ = -1;
    bool isValid_ = false;
    std::shared_ptr<MychainLib> mychain_lib_ptr_;
};


template<class M>
class MyMapIterable final : public MyMap<M> {
public:
    // No public constructors allowed. Use MyFactory to make instances instead.

    ~MyMapIterable() {}

    // One of the two primary API's (get and add) to allocate memory for MyIndexNodeM
    // Don't call MyIndexNodeM::make_instance() directly elsewhere.
    std::shared_ptr<MyIndexNodeM> get_index_node_by_index_key(
        std::string _index_key, const std::shared_ptr<MyBase>& _parent=nullptr) {
        std::shared_ptr<MyIndexNodeM> rv;
        if (index_nodes_.find(_index_key) != index_nodes_.end()) {
            rv = index_nodes_[_index_key].lock();
        }
        if (!rv) {
            std::unique_ptr<Key> my_index_key = this->base_key_->get_map_index_key(_index_key);
            std::string key = my_index_key->to_string();
            std::string buf;
            int ret = this->mychain_lib_ptr_->MyGetStorage(key, buf);
            if (ret != 0) {
                return nullptr;
            } else {
                rv = MyFactory<MyIndexNodeM>::make_instance(this->mychain_lib_ptr_, buf, std::move(my_index_key),
                    (_parent ? _parent : this->get_self_ptr()));

                uint16_t bitmap_size = rv->index_width_to_bitmap_size(get_index_width());
                // If the index node's bitmap is empty, fill it with zeros.
                if (rv->get_bitmap()->size() == 0) {
                    for (uint16_t i = 0; i<bitmap_size; i++) {
                        auto pbits = rv->get_bitmap()->append_element();
                        pbits->set_val(0);
                    }
                }
            }
            index_nodes_[_index_key] = rv;
        }
        return rv;
    }

    // One of the two primary API's (get and add) to allocate memory for MyIndexNodeM
    // Don't call MyIndexNodeM::make_instance() directly elsewhere.
    std::shared_ptr<MyIndexNodeM> add_index_node_by_index_key(
        std::string _index_key, const std::shared_ptr<MyBase>& _parent=nullptr) {
        std::shared_ptr<MyIndexNodeM> rv;
        std::unique_ptr<Key> my_index_key = this->base_key_->get_map_index_key(_index_key);
        rv = MyFactory<MyIndexNodeM>::make_instance(this->mychain_lib_ptr_, std::move(my_index_key),
            (_parent ? _parent : this->get_self_ptr()));

        // Fill the bitmap with zeros.
        uint16_t bitmap_size = rv->index_width_to_bitmap_size(get_index_width());
        for (uint16_t i = 0; i<bitmap_size; i++) {
            auto pbits = rv->get_bitmap()->append_element();
            pbits->set_val(0);
        }

        index_nodes_[_index_key] = rv;
        return rv;
    }

    using MyMap<M>::has_element;
    using MyMap<M>::get_element;

    std::shared_ptr<M> add_element(std::string _element_key, bool do_check [[gnu::unused]] = true) override {
        this->mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyMapIterable add_element: element_key %s", _element_key.c_str());
        if (has_element(_element_key))
            delete_element(_element_key);

        update_meta_data(hash_func_, index_width_, max_lcount_, MYBUFFER_MAPITERABLE_FORCE_UPDATE_META_BITMAP);
        add_element_key_to_index(_element_key);

        return MyMap<M>::add_element(_element_key, false/*do_check*/);
    }

    int delete_element(std::string _element_key, bool revert_on_non_existence = true) override {
        this->mychain_lib_ptr_->MYLOG(LogLevel::kInfo, "MyMapIterable delete_element: element_key %s", _element_key.c_str());
        int ret = MyMap<M>::delete_element(_element_key, revert_on_non_existence);
        if (ret) {
            return ret;
        }

        update_meta_data(hash_func_, index_width_, max_lcount_, MYBUFFER_MAPITERABLE_FORCE_UPDATE_META_BITMAP);
        ret = delete_element_key_in_index(_element_key);
        if (revert_on_non_existence && ret == MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS) {
            std::string msg = "Error: element was deleted in map, but failed to be deleted in index. element_key: " + _element_key;
            this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            this->mychain_lib_ptr_->MyRevert(msg);
        }
        return ret;
    }

    size_t size() {
        update_meta_data(hash_func_, index_width_, max_lcount_, true/*update_bitmap*/);

        size_t sum = 0;
        for (uint16_t i=0; i<get_index_width(); i++) {
            // Metadata Bitmap: access index nodes of level 0 according to the bitmap of metadata.
            bool exists = MyBitmapAccessor::get_bit(i, pmeta_->get_bitmap());
            if (!exists) {
                continue;
            }

            std::shared_ptr<MyIndexNodeM> pindex = get_index_node_by_index_key(std::to_string(i));
            if (pindex) {
                // tcount is at least as large as lcount in level 0.
                if (pindex->get_tcount() < pindex->get_lcount()) {
                    std::string msg = "tcount is less than lcount in map_iterable.size(), "
                                        " tcount: " + std::to_string(pindex->get_tcount()) +
                                        " lcount: " + std::to_string(pindex->get_lcount());
                    this->mychain_lib_ptr_->MYLOG(LogLevel::kWarning, msg);

                    pindex->set_tcount(pindex->get_lcount());
                }

                sum += pindex->get_tcount();
            }
        }
        return sum;
    }

    // Convenient function so that users don't need to write something like this:
    // MapKeyIterator<MyMapIterable<AccountM> > account_itr(paccounts);
    MapKeyIterator<MyMapIterable<M> > get_map_key_iterator() {
        return MapKeyIterator<MyMapIterable<M> >(std::static_pointer_cast<MyMapIterable>(this->get_self_ptr()),
            this->mychain_lib_ptr_);
    }

    HashFunc get_hash_function() const {
        return hash_func_;
    }

    uint16_t get_index_width() const {
        if (index_width_ == 0) {
            // Revert if invalid index_width
            std::string msg = "Invalid index width: 0";
            this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            this->mychain_lib_ptr_->MyRevert(msg);

            // To avoid the wrong error report in C++ pinpoint scan.
            return 1;
        }

        return index_width_;
    }

    uint16_t get_max_lcount() const {
        return max_lcount_;
    }

protected:
    MyMapIterable(
        const std::shared_ptr<MychainLib>& _mychain_lib_ptr,
        std::unique_ptr<Key> _key, std::shared_ptr<MyBase> _parent,
        const HashFunc _hash_func,
        const uint16_t _index_width,
        const uint16_t _max_lcount)
    : MyMap<M>(_mychain_lib_ptr, std::move(_key), _parent),
      hash_func_(_hash_func),
      index_width_(_index_width),
      max_lcount_(_max_lcount) {
        static_assert(std::is_base_of<MyBase, M>::value, "Type M must derive from MyTable");
    }

    std::shared_ptr<MyMapIterableMetaM> get_meta_data() const {
        return pmeta_;
    }

private:
    friend MyFactory<MyMapIterable<M> >;
    friend MapKeyIterator<MyMapIterable<M>>;

    /**
     * @brief get_naive_hash_value returns the naive hash value of the given
     * element key and level.
     *
     * @param _element_key
     * @param level
     *
     * @return hash value
    */
    int get_naive_hash_value(const std::string& _element_key, const int level) const {
        int hash_val = 0;
        int l = level + 1;
        for (uint32_t i = 0; i < _element_key.length(); i++) {
            hash_val += (uint8_t)_element_key[i] * l;
        }
        hash_val %= get_index_width();
        return hash_val;
    }

    /**
     * @brief get_city_hash_value returns the CityHash32%index_with value of
     * the given element key and level.
     *
     * @param _element_key
     * @param level
     *
     * @return hash value
    */
    int get_city_hash_value(const std::string& _element_key, int level) const {
        const std::string s = _element_key + std::to_string(level);
        const uint32_t h = CityHash32(s.c_str(), s.size());
        return h%get_index_width();
    }

    /**
     * @brief to_index_node_id returns a hash value according to
     * the given element key and level of index node.
     *
     * @param _element_key
     * @param level
     *
     * @return hash value.
    */
    int to_index_node_id(const std::string& _element_key, const int level) const {
        if (get_hash_function() == HashFunc::HashFunc_city) {
            if (level == 0) {
                return get_naive_hash_value(_element_key, level);
            } else {
                return get_city_hash_value(_element_key, level);
            }
        } else if (get_hash_function() == HashFunc::HashFunc_city_all) {
            return get_city_hash_value(_element_key, level);
        } else {
            std::string msg = "Unknow hash function: " + std::to_string(int(get_hash_function()));
            this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
            this->mychain_lib_ptr_->MyRevert(msg);
            return -1;
        }
    }

    /**
     * @brief get_index_node_id_and_key returns the index node id(hash value) and index key of
     * the given element key, according to the parent index node's index key and index level.
     *
     * @param element_key
     * @param level
     * @param parent_index_key
     *
     * @return pair of {index_node_id, index_key}
    */
    std::pair<int, std::string> get_index_node_id_and_key(
        const std::string& element_key, int level, const std::string& parent_index_key) {
        int index_node_id = to_index_node_id(element_key, level);
        if (parent_index_key.empty()) {
            return {index_node_id, std::to_string(index_node_id)};
        } else {
            return {index_node_id, Key::ConcatWithSlash(parent_index_key, std::to_string(index_node_id))};
        }
    }

    /**
     * @brief get_parent_index_node_id_and_key returns the index node id and
     * index key of parent level.
     *
     * @param element_key
     * @param level
     * @param child_index_key
     *
     * @return pair of {index_node_id_parent, index_key_parent}
    */
    std::pair<int, std::string> get_parent_index_node_id_and_key(
        const std::string& element_key, int level, const std::string& child_index_key) {
        int index_node_id = to_index_node_id(element_key, level);
        int index_node_id_parent = to_index_node_id(element_key, level-1);
        if (child_index_key.empty()) {
            return {index_node_id_parent, ""};
        } else {
            size_t pos = child_index_key.rfind(std::to_string(index_node_id));
            if (pos == std::string::npos || pos <= 1) {
                return {index_node_id_parent, ""};
            } else {
                return {index_node_id_parent, child_index_key.substr(0, pos-1)};
            }
        }
    }

    /**
     * @brief add_element_key_to_index adds an element key to the index structure in a recursive way.
     *
     * @param element_key The element key.
     * @return error code.
    */
    int add_element_key_to_index(const std::string& element_key) {
        std::shared_ptr<MyIndexNodeM> pindex;

        // Get the index_key of level 0.
        auto index_node_id_key = get_index_node_id_and_key(element_key, 0/*level*/, ""/*parent_index_key*/);
        std::string index_key = index_node_id_key.second;

        // Get index node of level 0, if not found, create it.
        pindex = get_index_node_by_index_key(index_key);
        if (!pindex) {
            pindex = add_index_node_by_index_key(index_key);
        }
        // Metadata Bitmap: update metadata's bitmap if the bitmap is of valid size.
        if (MyBitmapAccessor::check_bitmap_size(get_index_width(), pmeta_->get_bitmap())) {
            MyBitmapAccessor::set_bit(index_node_id_key.first, true, pmeta_->get_bitmap());
        }

        // tcount is at least as large as lcount in level 0.
        if (pindex->get_tcount() < pindex->get_lcount()) {
            std::string msg = "tcount is less than lcount in map_iterable.add_element(), "
                                " tcount: " + std::to_string(pindex->get_tcount()) +
                                " lcount: " + std::to_string(pindex->get_lcount());
            this->mychain_lib_ptr_->MYLOG(LogLevel::kWarning, msg);

            pindex->set_tcount(pindex->get_lcount());
        }

        // Add the given element key to the index node of level 0.
        int ret = add_element_key_to_index_node(element_key, pindex, index_key, 0/*level*/);

        // Increase tcount
        if (ret == MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
            pindex->set_tcount(pindex->get_tcount() + 1);
        }
        return ret;
    }

    /**
     * @brief delete_element_key_in_index deletes an element key in the index structure in a recursive way.
     *
     * @param element_key The element key.
     * @return error code.
    */
    int delete_element_key_in_index(const std::string& element_key) {
        std::shared_ptr<MyIndexNodeM> pindex;

        // Get the index_key of level 0.
        auto index_node_id_key = get_index_node_id_and_key(element_key, 0/*level*/, ""/*parent_index_key*/);
        uint16_t index_node_id = index_node_id_key.first;
        std::string index_key = index_node_id_key.second;

        // Get index node of level 0, if not found, create it.
        pindex = get_index_node_by_index_key(index_key);
        if (!pindex) {
            pindex = add_index_node_by_index_key(index_key);
        }

        // tcount is at least as large as lcount in level 0.
        if (pindex->get_tcount() < pindex->get_lcount()) {
            std::string msg = "tcount is less than lcount in map_iterable.delete_element(), "
                                " tcount: " + std::to_string(pindex->get_tcount()) +
                                " lcount: " + std::to_string(pindex->get_lcount());
            this->mychain_lib_ptr_->MYLOG(LogLevel::kWarning, msg);

            pindex->set_tcount(pindex->get_lcount());
        }

        // Delete the given element key in the index node of level 0.
        int ret = delete_element_key_in_index_node(element_key, pindex, index_node_id, index_key, 0/*level*/);

        // Decrease tcount
        if (ret == MapIterableErrors::MAP_ITERABLE_ERROR_NONE) {
            pindex->set_tcount(pindex->get_tcount() - 1);
        }
        return ret;
    }

    /**
     * @brief add_element_key_to_index_node adds an element key to an index
     * node, if the index node's key list if full, then add the element key to
     * the index node's sub-node (in next level).
     *
     * @param element_key   The element key.
     * @param pindex        The shared pointer to the index node.
     * @param index_key     The key of the index node(without the prefix, aka, the map_iterable's key).
     * @param level         The level of the index node, starts from 0.
     *
     * @return error code
    */
    int add_element_key_to_index_node(
        const std::string& element_key,
        const std::shared_ptr<MyIndexNodeM>& pindex,
        const std::string& index_key, const int level) {
            // Get the size of the index node.
            uint16_t lcount = pindex->get_lcount();

            if (lcount < get_max_lcount()) {
                // If there is enough space for the element key, add it.
                auto pkey = pindex->get_keys()->append_element();
                pkey->set_val(element_key);
                pindex->set_lcount(++lcount);
                return MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
            } else {
                // Else, check the bitmap and go to the next level.

                // If the size of the bitmap is not valid, revert
                uint16_t bitmap_size = pindex->index_width_to_bitmap_size(get_index_width());
                if (pindex->get_bitmap()->size() != bitmap_size) {
                    std::string msg = "Invalid bitmap size when add_element. got: "+std::to_string(pindex->get_bitmap()->size())+
                                      " want: "+std::to_string(bitmap_size);
                    this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                    this->mychain_lib_ptr_->MyRevert(msg);
                }

                // Get the index_node_id and index_key of the next level.
                int level_child = level + 1;
                auto index_node_id_key = get_index_node_id_and_key(element_key, level_child, index_key/*parent_index_key*/);
                int index_node_id_child = index_node_id_key.first;
                std::string index_key_child = index_node_id_key.second;
                std::shared_ptr<MyIndexNodeM> pindex_child;

                // If the bit at the expected position of the bitmap is 1,
                // get the child index node.
                if (pindex->get_bit_by_index_node_id(index_node_id_child, pindex->get_bitmap())) {
                    pindex_child = get_index_node_by_index_key(index_key_child, pindex);
                } else {
                    // Else, set the bit to 1 and add the child index node.
                    pindex->set_bit_by_index_node_id(index_node_id_child, true, pindex->get_bitmap());
                    pindex_child = get_index_node_by_index_key(index_key_child, pindex);
                    if (!pindex_child) {
                        pindex_child = add_index_node_by_index_key(index_key_child, pindex);
                    }
                }

                // Add the element key to the index node of next level.
                return add_element_key_to_index_node(element_key, pindex_child, index_key_child, level_child);
            }
    }

    /**
     *
     * @brief delete_element_key_in_index_node deletes an element key in an index node.
     * If the element key is not in the index node's key list, delete it in the
     * index node's sub-node (in next level).
     *
     * @param element_key   The element key.
     * @param pindex        The shared pointer to the index node.
     * @param index_node_id The id of the index node.
     * @param index_key     The key of the index node(without the prefix, aka, the map_iterable's key).
     * @param level         The level of the index node, starts from 0.
     *
     * @return error code
    */
    int delete_element_key_in_index_node(
        const std::string& element_key,
        const std::shared_ptr<MyIndexNodeM>& pindex,
        const uint16_t index_node_id,
        const std::string& index_key, const int level) {
            bool deleted = false;

            // Try to delete the element key in the index node's key list.
            int lcount = pindex->get_lcount();
            auto key_vector_ptr = pindex->get_keys();
            // TODO: use more efficient way to delete the element key.
            for (int i=0; i<lcount; i++) {
                auto pkey = key_vector_ptr->get_element(i);
                if (pkey->get_val() == element_key) {
                    key_vector_ptr->delete_element(i);
                    pindex->set_lcount(--lcount);
                    deleted = true;
                    break;
                }
            }

            if (deleted) {
                delete_empty_index_node_from_leaf(element_key, pindex, index_node_id, index_key, level);
                return MapIterableErrors::MAP_ITERABLE_ERROR_NONE;
            } else {
                // Else, check the bitmap and go to the next level.

                // If the size of the bitmap is not valid, revert
                uint16_t bitmap_size = pindex->index_width_to_bitmap_size(get_index_width());
                if (pindex->get_bitmap()->size() != bitmap_size) {
                    std::string msg = "Invalid bitmap size when delete_element. got: " + std::to_string(pindex->get_bitmap()->size()) +
                                      " want: " + std::to_string(bitmap_size);
                    this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                    this->mychain_lib_ptr_->MyRevert(msg);
                }

                // Get the index_node_id and index_key of the next level.
                int level_child = level + 1;
                auto index_node_id_key_child = get_index_node_id_and_key(
                    element_key, level_child, index_key/*parent_index_key*/);
                int index_node_id_child = index_node_id_key_child.first;
                std::string index_key_child = index_node_id_key_child.second;
                std::shared_ptr<MyIndexNodeM> pindex_child;

                // If the bit at the expected position of the bitmap is 1, get the index node.
                if (pindex->get_bit_by_index_node_id(index_node_id_child, pindex->get_bitmap())) {
                    pindex_child = get_index_node_by_index_key(index_key_child, pindex);
                } else {
                    // else, the element key not exists, return false.
                    return MapIterableErrors::MAP_ITERABLE_ERROR_ELEMENT_KEY_NOT_EXISTS;
                }

                // Delete the element key in the index node of next level.
                return delete_element_key_in_index_node(
                    element_key, pindex_child, index_node_id_child, index_key_child, level_child);
            }
    }

    /**
     * @brief delete_empty_index_node_from_leaf deletes empty index nodes recursively
     * from leaf to top level nodes.
     *
     * @param element_key
     * @param pindex
     * @param index_node_id
     * @param index_key
     * @param level
     *
     * @return void
     *
    */
    void delete_empty_index_node_from_leaf(
        const std::string& element_key,
        const std::shared_ptr<MyIndexNodeM>& pindex,
        const uint16_t index_node_id,
        const std::string& index_key, const int level) {
        // Check if the index node's keys and bitmap are both empty.
        bool empty = true;

        if (pindex->get_lcount() == 0) {
            // If the size of the bitmap is not valid, revert
            uint16_t bitmap_size = pindex->index_width_to_bitmap_size(get_index_width());
            if (pindex->get_bitmap()->size() != bitmap_size) {
                std::string msg = "Invalid bitmap size when delete_empty_index."
                                  "got: " + std::to_string(pindex->get_bitmap()->size()) +
                                  " want: " + std::to_string(bitmap_size);
                this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                this->mychain_lib_ptr_->MyRevert(msg);
            }

            // If there exist one bit that is not zero, it is not empty.
            for (uint16_t i = 0; i < get_index_width(); i++) {
                if (pindex->get_bit_by_index_node_id(i, pindex->get_bitmap())) {
                    empty = false;
                    break;
                }
            }
        } else {
            empty = false;
        }

        if (empty) {
            // We do not want to delete root nodes.
            if (level <= 0) {
                // Metadata Bitmap: if metadata's bitmap is of valid size,
                // delete the index node, mark the bit in bitmap of metadata as zero
                // else, just return
                if (MyBitmapAccessor::check_bitmap_size(get_index_width(), pmeta_->get_bitmap())) {
                    MyBitmapAccessor::set_bit(index_node_id, false, pmeta_->get_bitmap());
                }
                return;
            }

            // Get the index key of parent index node.
            auto index_node_id_and_key_parent = get_parent_index_node_id_and_key(element_key, level, index_key);
            int index_node_id_parent = index_node_id_and_key_parent.first;
            std::string index_key_parent = index_node_id_and_key_parent.second;
            if (index_key_parent.empty()) {
                std::string msg = "Failed to find parent index key, element_key: " + element_key +
                                  " level: " + std::to_string(level) + " index key: " + index_key;
                this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                this->mychain_lib_ptr_->MyRevert(msg);
            }

            // Get the parent index node.
            std::shared_ptr<MyIndexNodeM> pindex_parent = get_index_node_by_index_key(index_key_parent);

            // Set the corresponding bit to zero in parent node's bitmap.
            pindex_parent->set_bit_by_index_node_id(index_node_id, false, pindex_parent->get_bitmap());

            // Recursively delete empty parent nodes.
            delete_empty_index_node_from_leaf(element_key, pindex_parent, index_node_id_parent, index_key_parent, level-1/*level_parent*/);
        }
        return;
    }

protected:
    /**
     * @brief update_meta_data updates the pmeta_ member variable of MapIterable,
     * it will add a meta data node if necessary.
     *
     * The mybuffer key of the meta data node is composed like the key of index
     * node, with suffix `M`.
     * i.e.
     * map_iterable key: root_contract_Transfer.my_map_iterable
     * meta_data key v1 : root_contract_Transfer.my_map_iterable/M
     * meta_data key v2 : root_contract_Transfer\x03\x01/M
     *
     * @param _hash_func
     * @param _index_width
     * @param _max_lcount
     * @param update_bitmap
     *
     * @return if succeed
    */
    bool update_meta_data(
        const HashFunc _hash_func,
        const uint16_t _index_width,
        const uint16_t _max_lcount,
        bool update_bitmap) {

        if (pmeta_) {
            if (update_bitmap) {
                MyBitmapAccessor::update_bitmap<MyMapIterable<M>>(this, pmeta_->get_index_width(), pmeta_->get_bitmap());
            }
            return true;
        }

        std::shared_ptr<MyMapIterableMetaM> pmeta;

        // Read meta data from storage
        std::unique_ptr<Key> my_meta_key = this->base_key_->get_map_meta_key();
        std::string key = my_meta_key->to_string();
        std::string buf;
        int ret = this->mychain_lib_ptr_->MyGetStorage(key, buf);
        if (ret == 0) {
            // Meta data exists
            pmeta = MyFactory<MyMapIterableMetaM>::make_instance(
                this->mychain_lib_ptr_, buf, std::move(my_meta_key), this->get_parent_ptr());
            // use map_iterable's parent_ptr as pmeta's parent_ptr to avoid dead-lock when releasing.

            // Check if compatible.
            if (pmeta &&
                    ((pmeta->get_hash_function() != _hash_func) ||
                    (pmeta->get_index_width() != _index_width) ||
                    (pmeta->get_max_lcount() != _max_lcount))) {
                std::string msg = "MapIterable's attribute in schema is not compatible: Not Compatible With Existing Value!"
                                  " hash_func given: " + std::string(EnumNameHashFunc(_hash_func)) +
                                  " hash_func in storage: " + std::string(EnumNameHashFunc(pmeta->get_hash_function())) +
                                  " index_width given: " + std::to_string(_index_width) +
                                  " index_width in storage: " + std::to_string(pmeta->get_index_width()) +
                                  " max_lcount given: " + std::to_string(_max_lcount) +
                                  " max_lcount in storage: " + std::to_string(pmeta->get_max_lcount());
                this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                this->mychain_lib_ptr_->MyRevert(msg);
                return false;
            }

            if (update_bitmap) {
                MyBitmapAccessor::update_bitmap<MyMapIterable<M>>(this, _index_width, pmeta->get_bitmap());
            }

            pmeta_ = pmeta;
            return true;
        } else {
            // Meta data not exists

            std::vector<bool> bitmap_of_default(kMapIndexWidth);

            // Check if there is any index node of level 0;
            bool empty = true;
            for (uint32_t i = 0; i < kMapIndexWidth; i++) {
                std::shared_ptr<MyIndexNodeM> pindex = get_index_node_by_index_key(std::to_string(i));
                if (pindex) {
                    // The map is not empty
                    empty = false;

                    // Update tcount filed of index node of level 0
                    if(pindex->get_tcount() < pindex->get_lcount()) {
                        pindex->set_tcount(pindex->get_lcount());
                    }

                    // Update the bitmap
                    bitmap_of_default[i] = ((pindex->get_tcount() > 0) || (pindex->get_lcount() > 0));
                } else {
                    bitmap_of_default[i] = false;
                }
            }

            // If the map_iterable is not empty, check if the given meta data is equal to default values.
            if ((!empty) &&
                ((_hash_func != HashFunc::HashFunc_city) ||
                    (_index_width != kMapIndexWidth) ||
                    (_max_lcount != kMapIndexMaxLcount))) {
                std::string msg = "MapIterable's attribute in schema is not compatible: Not Default Value! hash_func: " +
                                  std::string(EnumNameHashFunc(_hash_func)) +
                                  " index_width: " + std::to_string(_index_width) + " max_lcount: " + std::to_string(_max_lcount);
                this->mychain_lib_ptr_->MYLOG(LogLevel::kError, msg);
                this->mychain_lib_ptr_->MyRevert(msg);
                return false;
            }

            // Add meta data
            pmeta = MyFactory<MyMapIterableMetaM>::make_instance(
                this->mychain_lib_ptr_, std::move(my_meta_key), this->get_parent_ptr());
            // use map_iterable's parent_ptr as pmeta's parent_ptr to avoid dead-lock when releasing.

            pmeta->set_hash_function(_hash_func);
            pmeta->set_index_width(_index_width);
            pmeta->set_max_lcount(_max_lcount);

            if (update_bitmap) {
                MyBitmapAccessor::update_bitmap(_index_width, bitmap_of_default, pmeta->get_bitmap());
            }

            pmeta_ = pmeta;
            return true;
        }
    }

private:
    std::map<std::string, std::weak_ptr<MyIndexNodeM> > index_nodes_;
    const HashFunc hash_func_;
    const uint16_t index_width_;
    const uint16_t max_lcount_;

    std::shared_ptr<MyMapIterableMetaM> pmeta_ = nullptr;
};

#endif // MYBUFFER_MYMAPITERABLE_H
