#ifndef MYBUFFER_WASM2DCVMLIB_H
#define MYBUFFER_WASM2DCVMLIB_H

#include <cstdarg>
#include <string>

#include "mybuffer/mychainlib.h"
#include "mychainlib/contract.h"  // Provided by mycdt

class Wasm2DcvmLib : public MychainLib {
 public:
  explicit Wasm2DcvmLib(mychain::Identity _dcId, uint32_t _dbId)
      : dcId_(_dcId), dbId_(_dbId) {}

  ~Wasm2DcvmLib() final = default;

  int MyPrint(const char *format, ...) override {
    char buf[LOGGER_BUFFER_SIZE];
    va_list args;
    va_start(args, format);
    vsnprintf(buf, LOGGER_BUFFER_SIZE, format, args);
    va_end(args);
    return mychain::println("%s", buf);
  }

  int MyPrintWithLevel(LogLevel level, const char* format, ...) override {
    if (level >= GetLogLevel()) {
      char buf[LOGGER_BUFFER_SIZE];
      va_list args;
      va_start(args, format);
      vsnprintf(buf, LOGGER_BUFFER_SIZE, (kLevelPrefixMap[level] + format).c_str(), args);
      va_end(args);
      return mychain::println("%s", buf);
    }

    return 0;
  }

  void MyRevert(const std::string &msg) override { mychain::Revert(msg); }

 protected:
  int MyGetStorageInternal(const std::string &key, std::string &val) override {
    return mychain::DCGetStorage(dcId_, dbId_, key, val);
  }

  int MySetStorageInternal(const std::string &key,
                           const std::string &val) override {
    return mychain::DCSetStorage(dcId_, dbId_, key, val);
  }

  int MyDelStorageInternal(const std::string &key) override {
    return mychain::DCDeleteStorage(dcId_, dbId_, key);
  }

 private:
  mychain::Identity dcId_;
  uint32_t dbId_;
};

#endif  // MYBUFFER_WASM2DCVMLIB_H