#ifndef MYBUFFER_WASMLIB_H
#define MYBUFFER_WASMLIB_H

#include "mybuffer/mychainlib.h"
#include "mychainlib/contract.h" // Provided by mycdt

#include <cstdarg>
#include <string>

class WasmLib : public MychainLib {
public:
    ~WasmLib() final = default;

    int MyPrint(const char* format, ...) override {
        char buf[LOGGER_BUFFER_SIZE];
        va_list args;
        va_start(args, format);
        vsnprintf(buf, LOGGER_BUFFER_SIZE, format, args);
        va_end(args);
        return mychain::println("%s", buf);
    }

    int MyPrintWithLevel(LogLevel level, const char* format, ...) override {
        if (level >= GetLogLevel()) {
            char buf[LOGGER_BUFFER_SIZE];
            va_list args;
            va_start(args, format);
            vsnprintf(buf, LOGGER_BUFFER_SIZE, (kLevelPrefixMap[level] + format).c_str(), args);
            va_end(args);
            return mychain::println("%s", buf);
        }

        return 0;
    }

    void MyRevert(const std::string& msg) override {
        mychain::Revert(msg);
    }

protected:
    int MyGetStorageInternal(const std::string &key, std::string &val) override {
        return mychain::GetStorage(key, val);
    }

    int MySetStorageInternal(const std::string &key, const std::string &val) override {
        return mychain::SetStorage(key, val);
    }

    int MyDelStorageInternal(const std::string &key) override {
        return mychain::DeleteStorage(key);
    }
};

#endif //MYBUFFER_WASMLIB_H