/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2018-12-28
 */

#pragma once

#include <string>
#include "types.h"

#ifdef __cplusplus
extern "C" {
#endif

namespace mychain {
namespace detail {

uint32_t DeployContract(const char* proposer_id_data,
		        const char* contract_id_data,
                        const char* code_data,
                        uint32_t code_len,
                        const char* auth_data,
                        uint32_t auth_len);

uint32_t UpdateContract(const char* contract_id_data,
                        const char* contract_code_data,
                        uint32_t contract_code_len);

uint32_t UpdateContractStatus(const char* contract_id_data,
                              uint32_t status);

int println(const char*, size_t);

int CheckAccount(const char* id_data, uint32_t id_len);

bool IsLocalTx();

int SetStorage(const char* key_data,
               uint32_t key_len,
               const char* val_data,
               uint32_t val_len);

int GetStorage(const char* key_data,
               uint32_t key_len,
               char* val_data,
               uint32_t* val_len_ptr);

int GetStorageSize(const char* key_data,
               uint32_t key_len,
               uint32_t* val_len_ptr);

int DeleteStorage(const char* key_data, uint32_t key_len);

uint64_t GetBlockNumber();

int GetBlockHash(uint64_t block_number,
                 char* block_hash_data,
                 uint32_t* block_hash_len_ptr);

uint64_t GetBlockTimeStamp();

int GetOrigin(char* origin_data, uint32_t* origin_data_len_ptr);

int GetAuthMap(const char* id_data,
               uint32_t id_len,
               char* auth_data,
               uint32_t* auth_len_ptr);

int GetAuthMapInCache(const char* id_data, uint32_t id_len,
                      char* auth_data, uint32_t* auth_len_ptr);

int GetBalance(const char* id_data, uint32_t id_len, uint64_t* val_ptr);

int GetCode(const char* id_data,
            uint32_t id_len,
            char* code_data,
            uint32_t* code_len_ptr);

int GetCodeHash(const char* id_data,
                uint32_t id_len,
                char* hash_data,
                uint32_t* hash_len_ptr);

int GetRecoverKey(const char* id_data,
                  uint32_t id_len,
                  char* recover_key_data,
                  uint32_t* recover_key_len_ptr);

int GetAccountStatus(const char* id_data, uint32_t id_len, uint32_t* status_ptr);

int GetTxHash(char* hash_data, uint32_t* hash_len_ptr);

uint64_t GetGas();

uint64_t GetValue();

uint32_t GetDataSize();

int GetData(char* data_ptr, uint32_t* data_len_ptr);

int TransferBalance(const char* to_data, uint32_t to_len, uint64_t balance);

[[noreturn]] int MyAbort(const char* exception_data, uint32_t exception_len);

int Result(const char* output_data, uint32_t output_len);

int Log(const char* user_custom_data,
        uint32_t user_custom_len,
        const char* topics_data,
        uint32_t topics_len);

int GetSelf(char* id_data, uint32_t* id_len_ptr);

int GetSender(char* id_data, uint32_t* id_len_ptr);

int VerifyRsa(const char* pk_data,
               uint32_t pk_len,
               const char* sig_data,
               uint32_t sig_len,
               const char* msg_data,
               uint32_t msg_len);

int VerifyRsa2(const char* pk_data,
                uint32_t pk_len,
                const char* sig_data,
                uint32_t sig_len,
                const char* msg_data,
                uint32_t msg_len,
                const uint16_t hash_type);

int VerifyMessageSM2(const char* pk_data,
                      uint32_t pk_len,
                      const char* sig_data,
                      uint32_t sig_len,
                      const char* msg_data,
                      uint32_t msg_len,
                      const uint16_t hash_type);

int Ecrecovery(const char* hash_data,
                uint32_t hash_len,
                const char* sig_data,
                uint32_t sig_len,
                char* id_data,
                uint32_t* id_len_ptr);

int Digest(const char* in_data,
            uint32_t in_len,
            const uint16_t type,
            char* out_data,
            uint32_t* out_len_ptr);

int Base64Encode(const char* in_data,
                  uint32_t in_len,
                  char* out_data,
                  uint32_t* out_len_ptr);

int Base64Decode(const char* in_data,
                  uint32_t in_len,
                  char* out_data,
                  uint32_t* out_len_ptr);
// zkproof//
int32_t RangeProofVerify(int32_t* ret_offset,
                         const char* proof,
                         uint32_t proof_len,
                         const char* pc_list,
                         uint32_t pc_list_len);

int32_t AddPedersenCommit(int32_t* ret_offset,
                          char* PC_result,
                          uint32_t* PC_result_len,
                          const char* PC_left,
                          uint32_t PC_left_len,
                          const char* PC_right,
                          uint32_t PC_right_len);

int32_t SubPedersenCommit(int32_t* ret_offset,
                          char* PC_result,
                          uint32_t* PC_result_len,
                          const char* PC_left,
                          uint32_t PC_left_len,
                          const char* PC_right,
                          uint32_t PC_right_len);

int32_t CalculatePedersenCommit(int32_t* ret_offset,
                                char* dst_pc,
                                uint32_t* dst_pc_len,
                                const char* src_pc,
                                uint32_t src_pc_len,
                                const char* positive,
                                uint32_t positive_len,
                                const char* negative,
                                uint32_t negative_len);

int32_t PedersenCommitEqualityVerify(int32_t* ret_offset,
                                     const char* positive,
                                     uint32_t positive_len,
                                     const char* negative,
                                     uint32_t negative_len);

uint32_t VerifyMessageECCK1(const char* pubkey,
                            uint32_t pubkey_len,
                            const char* sig,
                            uint32_t sig_len,
                            const char* msg,
                            uint32_t msg_len,
                            uint32_t hash_type);

uint32_t VerifyMessageECCR1(const char* pubkey,
                            uint32_t pubkey_len,
                            const char* sig,
                            uint32_t sig_len,
                            const char* msg,
                            uint32_t msg_len,
                            uint32_t hash_type);

int CreateContract(const char* src_id_data,
                   uint32_t src_id_len,
                   const char* new_id_data,
                   uint32_t new_id_len);

uint32_t GetTransaction(uint32_t field,
                          const char* input_data /*in*/,
                          uint32_t input_len, /*in*/
                          uint32_t* out_len_ptr /*out*/);

uint32_t GetConfidentialDeposit(uint32_t field,
                                 const char* input_data /*in*/,
                                 uint32_t input_len, /*in*/
                                 uint32_t* out_len_ptr /*out*/);

uint32_t GetRelatedTransactionListSize(const char* recipient_id,  /*in*/
                                       uint32_t recipient_id_len, /*in*/
                                       uint64_t deposit_flag /*in*/,
                                       uint64_t* count);

uint32_t GetRelatedTransactionList(const char* recipient_id,
                                     uint32_t recipient_id_len,
                                     uint64_t deposit_flag,
                                     uint64_t start_index,
                                     uint32_t size,
                                     uint32_t* out_len_ptr);

uint32_t ReadBuffer(char* out_data, uint32_t out_size);

uint32_t ReadBufferRef(char* out_data, uint32_t out_size);

int LiftedElgamalContractHomomorphicAdd(const char* first,
                                            uint32_t first_len,
                                            const char* second,
                                            uint32_t second_len,
                                            uint32_t* out_len);

int LiftedElgamalContractHomomorphicSub(const char* first,
                                            uint32_t first_len,
                                            const char* second,
                                            uint32_t second_len,
                                            uint32_t* out_len);

int LiftedElgamalScalarMutiply(const char* src,
                                   uint32_t src_len,
                                   uint64_t scalar,
                                   uint32_t* out_len);

int LiftedElgamalContractZeroCheckVerify(const char* ciphertext,
                                             uint32_t ciphertext_len,
                                             const char* proof,
                                             uint32_t proof_len,
                                             int* verify_result);

int LiftedElgamalContractRangeVerify(const char* ciphertext,
                                         uint32_t ciphertext_len,
                                         int per_value_bit_size,
                                         const char* public_key,
                                         uint32_t public_key_len,
                                         const char* proof,
                                         uint32_t proof_len,
                                         int* verify_result);

int BellmanSnarkVerify(const char* verification_key,
                                uint32_t verification_key_len,
                                const char* input,
                                uint32_t input_len,
                                const char* proof,
                                uint32_t proof_len);

uint32_t DCGetStorageSize(const char* dc_id_data,
		          uint32_t db_id,
			  const char* key_data,
			  uint32_t key_len,
			  uint32_t* len_offset);

uint32_t DCSetStorage(const char* dc_id_data,
		      uint32_t db_id,
	              const char* key_data,
	              uint32_t key_len,
		      const char* val_data,
		      uint32_t val_len);

uint32_t DCDeleteStorage(const char* dc_id_data,
		         uint32_t db_id,
	                 const char* key_data,
	                 uint32_t key_len);

uint32_t DCSetAcl(const char* contract_id,
                  uint32_t db_id,
                  const char* objects,
                  uint32_t objects_len);

// API for Grayscale
uint32_t GrayscaleDeployContract(const char* contract_id_data,
                                 const char* contract_code_data,
				 uint32_t contract_code_len);

uint32_t GrayscaleVerification(const char* contract_id_data);

uint32_t GrayscaleVersionSwitchBack(const char* contract_id_data);

uint32_t GrayscaleUpdateContract(const char* contract_id_data);

// 0: Unknown
// 1: SHA256
// 2: SM3
// 3: KECCAK256
uint32_t GetDigestType();

}  // namespace detail
}  // namespace mychain

#ifdef __cplusplus
}
#endif
