/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2018-12-28
 */

#pragma once
#include <boost/fusion/include/std_tuple.hpp>
#include <boost/mp11/tuple.hpp>
#include <boost/preprocessor/seq.hpp>
#include <boost/preprocessor/stringize.hpp>

#include <mychainlib/types.h>

namespace mychain {

extern "C" {
namespace detail {

int SetReturnValue(const char* in_data, uint32_t in_len);
int ReadInterfaceName(char* const name_data, uint32_t* name_len_ptr);
int ReadInterfaceParams(char* const params_data, uint32_t* params_len_ptr);
uint32_t ReadInterfaceNameSize();
uint32_t ReadInterfaceParamsSize();

} // namespace detail
} // extern "C"

template <typename C,
          typename R,
          typename T,
          typename... Args>
inline std::string execute_interface(R (T::*func)(Args...), const std::string& name) {
    uint32_t ps = detail::ReadInterfaceParamsSize();
    std::string params(ps, 0);
    detail::ReadInterfaceParams(params.data(), &ps);
    std::tuple<std::decay_t<Args>...> args;
    datastream<const char*> ds(params.data(), params.size(),
                               "PARAMS_NOT_MATCH: " + name + " ");
    ds >> args;
    Require(0 == ds.remaining(), "PARAMS_NOT_MATCH: " + name);

    C inst;
    auto f2 = [&](auto... a) -> R { return ((&inst)->*func)(a...); };

    if constexpr (std::is_same_v<R, void>) {
        boost::mp11::tuple_apply(f2, args);
        return std::string();
    } else {
        R ret = boost::mp11::tuple_apply(f2, args);
        return pack(ret);
    }
}

}  // namespace mychain

#define INTERFACE_EXPORT_INTERNAL(r, OP, elem)                           \
    if (name == BOOST_PP_STRINGIZE(elem)) {                              \
        auto value = mychain::execute_interface<OP>(&OP::elem, name);    \
        mychain::detail::SetReturnValue(value.data(), value.size());     \
        return mychain::EXECUTE_STATUS::SUCCESS;                         \
    }

#define INTERFACE_EXPORT(TYPE, MEMBERS)                                  \
    extern "C" int apply() {                                             \
        uint32_t ns = mychain::detail::ReadInterfaceNameSize();          \
        std::string name(ns, 0);                                         \
        mychain::detail::ReadInterfaceName(name.data(), &ns);            \
        BOOST_PP_SEQ_FOR_EACH(INTERFACE_EXPORT_INTERNAL, TYPE, MEMBERS); \
        mychain::Revert("METHOD_NOT_FOUND: " + name);                    \
    }
