/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2019-01-11
 */
#pragma once

#include <boost/preprocessor/seq/enum.hpp>
#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/seq/size.hpp>
#include <boost/preprocessor/stringize.hpp>
#include "datastream.hpp"

#define REFLECT_MEMBER_OP(r, OP, elem) OP t.elem

#define SERIALIZE(TYPE, MEMBERS)                                         \
    template <typename D>                                                \
    friend mychain::datastream<D>& operator<<(mychain::datastream<D>& ds, const TYPE& t) { \
        return ds BOOST_PP_SEQ_FOR_EACH(REFLECT_MEMBER_OP, <<, MEMBERS); \
    }                                                                    \
    template <typename D>                                                \
    friend mychain::datastream<D>& operator>>(mychain::datastream<D>& ds, TYPE& t) {       \
        return ds BOOST_PP_SEQ_FOR_EACH(REFLECT_MEMBER_OP, >>, MEMBERS); \
    }

#define SERIALIZE_DERIVED(TYPE, BASE, MEMBERS)                           \
    template <typename D>                                                \
    friend mychain::datastream<D>& operator<<(mychain::datastream<D>& ds, const TYPE& t) { \
        ds << static_cast<const BASE&>(t);                               \
        return ds BOOST_PP_SEQ_FOR_EACH(REFLECT_MEMBER_OP, <<, MEMBERS); \
    }                                                                    \
    template <typename D>                                                \
    friend mychain::datastream<D>& operator>>(mychain::datastream<D>& ds, TYPE& t) {       \
        ds >> static_cast<BASE&>(t);                                     \
        return ds BOOST_PP_SEQ_FOR_EACH(REFLECT_MEMBER_OP, >>, MEMBERS); \
    }
