/**
 * Copyright 2018 Ant Financial Group. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author : dongwei.ldw@antfin.com
 * Date : 2019-01-15
 */
#pragma once

#include <boost/preprocessor/seq/enum.hpp>
#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/seq/size.hpp>
#include <boost/preprocessor/stringize.hpp>
#include "types.h"

namespace mychain {
int SetStorage(const std::string& key, const std::string& value);

int GetStorage(const std::string& key, std::string& value);

int DeleteStorage(const std::string& key);

template <typename T>
int SetStorage(const std::string& key, const T& value) {
    int ret = SetStorage(key, pack(value));
    return ret == 0 ? 0 : 1;
}

template <typename T>
int GetStorage(const std::string& key, T& value) {
    std::string value_bytes;
    int ret = GetStorage(key, value_bytes);
    if (0 == ret) {
        value = unpack<T>(value_bytes, "MYCHAINLIB_ERROR: GetStorage ");
    }
    return ret == 0 ? 0 : 1;
}

#define NOT_DAMP 0

#define GET_ELEM_KEY(elem)                                    \
    const char* pname = BOOST_PP_STRINGIZE(elem);             \
    std::string key = {NOT_DAMP};                             \
    size_t elem_size = strlen(pname);                         \
    MyAssert(elem_size <= 255,                                \
             "MYCHAINLIB_ERROR: member elem size too large, " \
             "should be less than 255 bytes");                \
    key.push_back(elem_size);                                 \
    key.insert(key.end(), pname, pname + elem_size);

#define GET_STORAGE_MEMBER(r, Class, elem) \
    {                                      \
        elem = decltype(elem)();           \
        GET_ELEM_KEY(elem)                 \
        decltype(elem) value;              \
        if (0 == GetStorage(key, value)) { \
            elem = value;                  \
        }                                  \
    }

#define SET_STORAGE_MEMBER(r, Class, elem) \
    {                                      \
        GET_ELEM_KEY(elem)                 \
        SetStorage(key, elem);             \
    }

#define STORAGE(TYPE, MEMBERS)                                    \
public:                                                           \
    void MyChainLoadValue() override {                            \
        BOOST_PP_SEQ_FOR_EACH(GET_STORAGE_MEMBER, TYPE, MEMBERS); \
    }                                                             \
    void MyChainSaveValue() override {                            \
        BOOST_PP_SEQ_FOR_EACH(SET_STORAGE_MEMBER, TYPE, MEMBERS); \
    }                                                             \
    TYPE() { MyChainLoadValue(); }                                \
    virtual ~TYPE() { MyChainSaveValue(); }
};  // namespace mychain
