
#ifndef VM_UTIL_H
#define VM_UTIL_H

#include <vector>
#include <string>
#include <map>
#include <byteswap.h>
namespace mychain {
using bytes = std::vector<uint8_t>;

enum VMType : uint32_t {
    NATIVE = 0,
    EVM = 1,
    WASM = 2,
    DCVM = 4,
};

static uint64_t Bytes2Int64(const uint8_t* input) {
    uint64_t r = 0;
    unsigned char* p = reinterpret_cast<unsigned char*>(&r);
    p[0] = input[31];
    p[1] = input[30];
    p[2] = input[29];
    p[3] = input[28];
    p[4] = input[27];
    p[5] = input[26];
    p[6] = input[25];
    p[7] = input[24];
    return r;
}

class VMUtil {
public:
    static void ConstructUpdateData(VMType vm_type, bytes& code, bytes& data) {
        data.clear();

        // set evm type
        SetIdentityValue(vm_type, data);

        // assign code
        data.insert(data.end(), code.begin(), code.end());
    }

    static bool ReturnBoolValue(bytes& result, uint64_t parameter_th, bool& value) {
        if (parameter_th * 32 + 32 > result.size())
            return false;
        value = (result[32 * parameter_th + 31] != 0);
        return value;
    }

    static bool ReturnIntValue(bytes& result, uint64_t parameter_th, int64_t& value) {
        if (parameter_th * 32 + 32 > result.size())
            return false;
        value = static_cast<int64_t>(Bytes2Int64(parameter_th * 32 + result.data()));
        return true;
    }

    static bool ReturnUintValue(bytes& result, uint64_t parameter_th, uint64_t& value) {
        if (parameter_th * 32 + 32 > result.size())
            return false;
        value = Bytes2Int64(parameter_th * 32 + result.data());
        return true;
    }

    static bool ReturnIdentityValue(bytes& result, uint64_t parameter_th, Identity& value) {
        std::string tmp;
        if (parameter_th * 32 + 32 > result.size())
            return false;
        tmp.assign(reinterpret_cast<char*>(result.data() + parameter_th * 32), 32);
        value = Identity(tmp);
        return true;
    }

    static bool ReturnBytesNValue(bytes& result, uint64_t parameter_th, bytes& value) {
        if (parameter_th * 32 + 32 > result.size())
            return false;

        value.assign(result.data() + parameter_th * 32,
                     result.data() + parameter_th * 32 + 32);
        return true;
    }

    static bool ReturnStringValue(bytes& result, uint64_t parameter_th, std::string& value) {
        if (parameter_th * 32 + 32 > result.size()) {
            return false;
        }
        uint64_t offset = Bytes2Int64(result.data() + parameter_th * 32);
        if (offset + 32 > result.size()) {
            return false;
        }
        uint64_t count = Bytes2Int64(result.data() + offset);
        if (offset + 32 + count > result.size()) {
            return false;
        }
        value.assign(reinterpret_cast<char*>(result.data()) + offset + 32, count);

        return true;
    }

    static bool ReturnBytesValue(bytes& result, uint64_t parameter_th, bytes& value) {
        if (parameter_th * 32 + 32 > result.size()) {
            return false;
        }
        uint64_t offset = Bytes2Int64(result.data() + parameter_th * 32);
        if (offset + 32 > result.size()) {
            return false;
        }
        uint64_t count = Bytes2Int64(result.data() + offset);
        if (offset + 32 + count > result.size()) {
            return false;
        }
        value.assign(result.data() + offset + 32, result.data() + offset + 32 + count);

        return true;
    }

    static bool ReturnBytesNArrayValue(bytes& result,
                                uint64_t parameter_th,
                                std::vector<bytes>& re) {
        if (parameter_th * 32 + 32 > result.size()) {
            return false;
        }
        uint64_t offset = Bytes2Int64(parameter_th * 32 + result.data());
        if (offset + 32 > result.size()) {
            return false;
        }
        uint64_t count = Bytes2Int64(result.data() + offset);
        for (uint64_t i = 0; i < count; i++) {
            if (offset + (i + 1) * 32 + 32 > result.size()) {
                return false;
            }
            bytes tmp;
            tmp.assign(result.data() + offset + (i + 1) * 32,
                       result.data() + offset + (i + 1) * 32 + 32);
            re.push_back(tmp);
        }
        return true;
    }

    static bool ReturnIntArrayValue(bytes& result,
                             uint64_t parameter_th,
                             std::vector<int64_t>& re) {
        /*  try {
             bytesConstRef ref(&result);
             u256 offset = (u256)h256(ref.cropped(parameter_th * 32, 32).toBytes());
             u256 count = (u256)h256(ref.cropped((size_t)offset, 32).toBytes());
             for (int i = 0; i < count; i++) {
                 auto value = (u256)h256(ref.cropped((size_t)offset + (i + 1) * 32,
         32).toBytes()); re.push_back(static_cast<uint64_t>(value));
             }
         } catch (const std::exception &e) {
             return false;
         } */

        if (parameter_th * 32 + 32 > result.size())
            return false;
        uint64_t offset = Bytes2Int64(result.data() + parameter_th * 32);
        if (offset + 32 > result.size())
            return false;
        uint64_t count = Bytes2Int64(result.data() + offset);
        for (size_t i = 0; i < count; i++) {
            if (offset + (i + 1) * 32 + 32 > result.size())
                return false;
            int64_t v = Bytes2Int64(result.data() + offset + (i + 1) * 32);
            re.push_back(v);
        }

        return true;
    }

    static bool ReturnBoolArrayValue(bytes& result,
                              uint64_t parameter_th,
                              std::vector<bool>& re) {
        // try {
        //     bytesConstRef ref(&result);
        //     u256 offset = (u256)h256(ref.cropped(parameter_th * 32, 32).toBytes());
        //     u256 count = (u256)h256(ref.cropped((size_t)offset, 32).toBytes());
        //     for (int i = 0; i < count; i++) {
        //         u256 value = (u256)h256(ref.cropped((size_t)offset + (i + 1) * 32,
        //         32).toBytes()); bool result = value ? true : false;
        //         re.push_back(result);
        //     }
        // } catch (const std::exception &e) {
        //     return false;
        // }

        if (parameter_th * 32 + 32 > result.size())
            return false;
        uint64_t offset = Bytes2Int64(result.data() + parameter_th * 32);
        if (offset + 32 > result.size())
            return false;
        uint64_t count = Bytes2Int64(result.data() + offset);
        for (size_t i = 0; i < count; i++) {
            if (offset + (i + 1) * 32 + 32 > result.size())
                return false;
            int64_t v = Bytes2Int64(result.data() + offset + (i + 1) * 32);
            re.push_back(v != 0);
        }

        return true;
    }

    static bool ReturnUintArrayValue(bytes& result,
                              uint64_t parameter_th,
                              std::vector<uint64_t>& re) {
        if (parameter_th * 32 + 32 > result.size())
            return false;
        uint64_t offset = Bytes2Int64(result.data() + parameter_th * 32);
        if (offset + 32 > result.size())
            return false;
        uint64_t count = Bytes2Int64(result.data() + offset);
        for (size_t i = 0; i < count; i++) {
            if (offset + (i + 1) * 32 + 32 > result.size())
                return false;
            int64_t v = Bytes2Int64(result.data() + offset + (i + 1) * 32);
            re.push_back(v);
        }
        return true;
    }

    static void SetIdentityValue(VMType vm_type, bytes& data) {
        if (vm_type == VMType::EVM) {
            data.push_back(1);
        } else if (vm_type == VMType::NATIVE) {
            data.push_back(255);
        } else if (vm_type == VMType::WASM) {
            data.push_back(2);
        }
    }

    static bool decodeVMCode(const bytes& code, VMType& type) {
        if (code.empty()) {
            return false;
        }

        if (code[0] == 1) {
            type = VMType::EVM;
        } else if (code[0] == 255) {
            type = VMType::NATIVE;
        } else if (code[0] == 2) {
            type = VMType::WASM;
        } else {
            return false;
        }

        return true;
    }
};

}  // namespace mychain
#endif
